"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .activity_utils import extract_tenant_id
from .retry import RetryOptions, retry
from .timer import Timeout

__all__ = ["extract_tenant_id", "retry", "Timeout", "RetryOptions"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import asyncio
import importlib.metadata
from contextlib import AsyncExitStack, asynccontextmanager
from logging import Logger
from pathlib import Path
from types import SimpleNamespace
from typing import (
    Annotated,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    Optional,
    TypedDict,
    Union,
    Unpack,
    cast,
)

import uvicorn
from fastapi import FastAPI, Request, Response
from fastapi.staticfiles import StaticFiles
from microsoft.teams.api import (
    Activity,
    ActivityParams,
    ActivityTypeAdapter,
    ApiClient,
    ConversationReference,
    SentActivity,
    TokenProtocol,
)
from microsoft.teams.api.auth.credentials import Credentials
from microsoft.teams.apps.http_stream import HttpStream
from microsoft.teams.common.http import Client, ClientOptions, Token
from microsoft.teams.common.logging import ConsoleLogger
from pydantic import BaseModel, ValidationError
from starlette.applications import Starlette
from starlette.types import Lifespan

from .auth import create_jwt_validation_middleware
from .events import ActivityEvent, ErrorEvent
from .plugins import (
    DependencyMetadata,
    EventMetadata,
    LoggerDependencyOptions,
    PluginActivityResponseEvent,
    PluginErrorEvent,
    PluginStartEvent,
    Sender,
    StreamerProtocol,
)
from .plugins.metadata import Plugin

version = importlib.metadata.version("microsoft-teams-apps")


class HttpPluginOptions(TypedDict, total=False):
    """Options for configuring the HTTP plugin."""

    logger: Logger
    skip_auth: bool
    server_factory: Callable[[FastAPI], uvicorn.Server]


@Plugin(name="http", version=version, description="the default plugin for sending/receiving activities")
class HttpPlugin(Sender):
    """
    Basic HTTP plugin that provides a FastAPI server for Teams activities.
    """

    logger: Annotated[Logger, LoggerDependencyOptions()]
    credentials: Annotated[Optional[Credentials], DependencyMetadata(optional=True)]

    on_error_event: Annotated[Callable[[ErrorEvent], None], EventMetadata(name="error")]
    on_activity_event: Annotated[Callable[[ActivityEvent], None], EventMetadata(name="activity")]

    client: Annotated[Client, DependencyMetadata()]

    bot_token: Annotated[Optional[Callable[[], Token]], DependencyMetadata(optional=True)]

    lifespans: list[Lifespan[Starlette]] = []

    def __init__(self, **options: Unpack[HttpPluginOptions]):
        """
        Args:
            logger: Optional logger.
            skip_auth: Whether to skip JWT validation.
            server_factory: Optional function that takes an ASGI app
                and returns a configured `uvicorn.Server`.
            Example:
                ```python
                def custom_server_factory(app: FastAPI) -> uvicorn.Server:
                    return uvicorn.Server(config=uvicorn.Config(app, host="0.0.0.0", port=8000))


                http_plugin = HttpPlugin(app_id="your-app-id", server_factory=custom_server_factory)
                ```
        """
        super().__init__()
        self.logger = options.get("logger") or ConsoleLogger().create_logger("@teams/http-plugin")
        self._port: Optional[int] = None
        self._skip_auth: bool = options.get("skip_auth", False)
        self._server: Optional[uvicorn.Server] = None
        self._on_ready_callback: Optional[Callable[[], Awaitable[None]]] = None
        self._on_stopped_callback: Optional[Callable[[], Awaitable[None]]] = None

        # Storage for pending HTTP responses by activity ID
        self.pending: Dict[str, asyncio.Future[Any]] = {}

        # Setup FastAPI app with lifespan
        @asynccontextmanager
        async def default_lifespan(_app: Starlette) -> AsyncGenerator[None, None]:
            # Startup
            self.logger.info(f"listening on port {self._port} 🚀")
            if self._on_ready_callback:
                await self._on_ready_callback()
            yield
            # Shutdown
            self.logger.info("Server shutting down")
            if self._on_stopped_callback:
                await self._on_stopped_callback()

        @asynccontextmanager
        async def combined_lifespan(app: Starlette):
            async with AsyncExitStack() as stack:
                lifespans = self.lifespans.copy()
                lifespans.append(default_lifespan)
                for lifespan in lifespans:
                    await stack.enter_async_context(lifespan(app))
                yield

        self.app = FastAPI(lifespan=combined_lifespan)

        # Create uvicorn server if user provides custom factory method
        server_factory = options.get("server_factory")
        if server_factory:
            self._server = server_factory(self.app)
            if self._server.config.app is not self.app:
                raise ValueError(
                    "server_factory must return a uvicorn.Server configured with the provided FastAPI app instance."
                )

        # Expose FastAPI routing methods (like TypeScript exposes Express methods)
        self.get = self.app.get
        self.post = self.app.post
        self.put = self.app.put
        self.patch = self.app.patch
        self.delete = self.app.delete
        self.middleware = self.app.middleware

        # Setup routes and error handlers
        self._setup_routes()

    @property
    def on_ready_callback(self) -> Optional[Callable[[], Awaitable[None]]]:
        """Callback to call when HTTP server is ready."""
        return self._on_ready_callback

    @on_ready_callback.setter
    def on_ready_callback(self, callback: Optional[Callable[[], Awaitable[None]]]) -> None:
        """Set callback to call when HTTP server is ready."""
        self._on_ready_callback = callback

    @property
    def on_stopped_callback(self) -> Optional[Callable[[], Awaitable[None]]]:
        """Callback to call when HTTP server is stopped."""
        return self._on_stopped_callback

    @on_stopped_callback.setter
    def on_stopped_callback(self, callback: Optional[Callable[[], Awaitable[None]]]) -> None:
        """Set callback to call when HTTP server is stopped."""
        self._on_stopped_callback = callback

    async def on_init(self) -> None:
        """
        Initialize the HTTP plugin when the app starts.
        This adds JWT validation middleware unless `skip_auth` is True.
        """

        # Add JWT validation middleware
        app_id = getattr(self.credentials, "client_id", None)
        if app_id and not self._skip_auth:
            jwt_middleware = create_jwt_validation_middleware(
                app_id=app_id, logger=self.logger, paths=["/api/messages"]
            )
            self.app.middleware("http")(jwt_middleware)

    async def on_start(self, event: PluginStartEvent) -> None:
        """Start the HTTP server."""
        self._port = event.port

        try:
            if self._server and self._server.config.port != self._port:
                self.logger.warning(
                    "Using port configured by server factory: %d, but plugin start event has port %d.",
                    self._server.config.port,
                    self._port,
                )
                self._port = self._server.config.port
            else:
                config = uvicorn.Config(app=self.app, host="0.0.0.0", port=self._port, log_level="info")
                self._server = uvicorn.Server(config)

            self.logger.info("Starting HTTP server on port %d", self._port)

            # The lifespan handler will call the callback when the server is ready
            await self._server.serve()

        except OSError as error:
            # Handle port in use, permission errors, etc.
            self.logger.error("Server startup failed: %s", error)
            raise
        except Exception as error:
            self.logger.error("Failed to start server: %s", error)
            raise

    async def on_stop(self) -> None:
        """Stop the HTTP server."""
        if self._server:
            self.logger.info("Stopping HTTP server")
            self._server.should_exit = True

    async def on_activity_response(self, event: PluginActivityResponseEvent) -> None:
        """
        Complete a pending activity response.

        This is called when the App finishes processing an activity
        and is ready to send the HTTP response back.

        Args:
            activity_id: The ID of the activity to respond to
            response_data: The response data to send back
            plugin: The plugin that sent the response
        """
        future = self.pending.get(event.activity.id)
        if future and not future.done():
            future.set_result(event.response)

        else:
            self.logger.warning(f"No pending future found for activity {event.activity.id}")

    async def on_error(self, event: PluginErrorEvent) -> None:
        """
        Handle errors from the App.

        Args:
            error: The error that occurred
            activity_id: The ID of the activity that failed (if applicable)
            plugin: The plugin that caused the error (if applicable)
        """
        activity_id = event.activity.id if event.activity else None
        error = event.error
        if activity_id:
            future = self.pending.get(activity_id)
            if future and not future.done():
                future.set_exception(error)
                self.logger.error(f"Activity {activity_id} failed: {error}")
            else:
                self.logger.warning(f"No pending future found for activity {activity_id} (error: {error})")
        else:
            self.logger.error(f"Plugin error: {error}")

    async def send(self, activity: ActivityParams, ref: ConversationReference) -> SentActivity:
        api = ApiClient(service_url=ref.service_url, options=self.client.clone(ClientOptions(token=self.bot_token)))

        activity.from_ = ref.bot
        activity.conversation = ref.conversation

        if hasattr(activity, "id") and activity.id:
            res = await api.conversations.activities(ref.conversation.id).update(activity.id, activity)
            return SentActivity.merge(activity, res)

        res = await api.conversations.activities(ref.conversation.id).create(activity)
        return SentActivity.merge(activity, res)

    async def _process_activity(self, activity: Activity, activity_id: str, token: TokenProtocol) -> None:
        """
        Process an activity via the registered handler.

        Args:
            activity: The Teams activity data
            token: The authorization token (if any)
            activity_id: The activity ID for response coordination
        """
        try:
            event = ActivityEvent(activity=activity, sender=self, token=token)
            if asyncio.iscoroutinefunction(self.on_activity_event):
                await self.on_activity_event(event)
            else:
                self.on_activity_event(event)
        except Exception as error:
            # Complete with error
            await self.on_error(PluginErrorEvent(sender=self, error=error, activity=activity))

    async def _handle_activity_request(self, request: Request) -> Any:
        """
        Process the activity request and coordinate response.

        Args:
            request: The FastAPI request object (token is in request.state.validated_token)

        Returns:
            The activity processing result
        """
        # Parse activity data
        body = await request.json()

        # Get validated token from middleware (if present - will be missing if skip_auth is True)
        if hasattr(request.state, "validated_token") and request.state.validated_token:
            token = request.state.validated_token
        else:
            token = cast(
                TokenProtocol,
                SimpleNamespace(
                    app_id="",
                    app_display_name="",
                    tenant_id="",
                    service_url=body.get("serviceUrl", ""),
                    from_="azure",
                    from_id="",
                    is_expired=lambda: False,
                ),
            )

        activity_type = body.get("type", "unknown")
        activity_id = body.get("id", "unknown")

        self.logger.debug(f"Received activity: {activity_type} (ID: {activity_id})")

        # Create Future for async response coordination
        response_future = asyncio.get_event_loop().create_future()
        self.pending[activity_id] = response_future

        try:
            activity = ActivityTypeAdapter.validate_python(body)
        except ValidationError as e:
            self.logger.error(e.errors())
            raise

        # Fire activity processing via callback
        try:
            # Call the activity handler asynchronously
            self.logger.debug(f"Processing activity {activity_id} via handler...")
            asyncio.create_task(self._process_activity(activity, activity_id, token))
        except Exception as error:
            self.logger.error(f"Failed to start activity processing: {error}")
            response_future.set_exception(error)

        # Wait for the activity processing to complete
        result = await response_future

        # Clean up
        if activity_id in self.pending:
            del self.pending[activity_id]

        return result

    def _handle_activity_response(self, response: Response, result: Any) -> Union[Response, Dict[str, object]]:
        """
        Handle the activity response formatting.

        Args:
            response: The FastAPI response object
            result: The result from activity processing

        Returns:
            The formatted response
        """
        status_code: Optional[int] = None
        body: Optional[Dict[str, Any]] = None
        resp_dict: Optional[Dict[str, Any]] = None
        if isinstance(result, dict):
            resp_dict = cast(Dict[str, Any], result)
        elif isinstance(result, BaseModel):
            resp_dict = result.model_dump(exclude_none=True)

        # if resp_dict has status set it
        if resp_dict and "status" in resp_dict:
            status_code = resp_dict.get("status")

        if resp_dict and "body" in resp_dict:
            body = resp_dict.get("body", None)

        if status_code is not None:
            response.status_code = status_code

        if body is not None:
            self.logger.debug(f"Returning body {body}")
            return body
        self.logger.debug("Returning empty body")
        return response

    async def on_activity_request(self, request: Request, response: Response) -> Any:
        """Handle incoming Teams activity."""
        # Process the activity (token validation handled by middleware)
        result = await self._handle_activity_request(request)
        return self._handle_activity_response(response, result)

    def _setup_routes(self) -> None:
        """Setup FastAPI routes."""

        self.app.post("/api/messages")(self.on_activity_request)

        async def health_check() -> Dict[str, Any]:
            """Basic health check endpoint."""
            return {"status": "healthy", "port": self._port}

        self.app.get("/")(health_check)

    def create_stream(self, ref: ConversationReference) -> StreamerProtocol:
        """Create a new streaming instance."""

        api = ApiClient(ref.service_url, self.client.clone(ClientOptions(token=self.bot_token)))

        return HttpStream(api, ref, self.logger)

    def mount(self, name: str, dir_path: Path | str, page_path: Optional[str] = None) -> None:
        """
        Serve a static page at the given path.

        Args:
            name: The name of the page (used in URL)
            page_path: The path to the static HTML file
        """
        self.app.mount(page_path or f"/{name}", StaticFiles(directory=dir_path, check_dir=True, html=True), name=name)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Any, Awaitable, Callable, Optional, TypeVar, Union

from microsoft.teams.api import ActivityBase, InvokeResponse, InvokeResponseBody

from .activity_context import ActivityContext

# Type variables for generic activity and response types
ActivityT = TypeVar("ActivityT", bound=ActivityBase)
ResponseT = TypeVar("ResponseT", bound=InvokeResponseBody)

# Basic handler types that
BasicHandler = Callable[[ActivityContext[ActivityT]], Awaitable[Optional[Any]]]
BasicHandlerDecorator = Callable[[BasicHandler[ActivityT]], BasicHandler[ActivityT]]

# Invoke handler types that
InvokeHandler = Callable[
    [ActivityContext[ActivityT]],
    Awaitable[Union[InvokeResponse[ResponseT], ResponseT]],
]
InvokeHandlerDecorator = Callable[[InvokeHandler[ActivityT, ResponseT]], InvokeHandler[ActivityT, ResponseT]]

# Special case for handlers that return None in invoke scenarios
VoidInvokeHandler = Callable[[ActivityContext[ActivityT]], Awaitable[Union[InvokeResponse[None], None]]]
VoidInvokeHandlerDecorator = Callable[[VoidInvokeHandler[ActivityT]], VoidInvokeHandler[ActivityT]]

# Union types for overloaded methods
BasicHandlerUnion = Union[BasicHandlerDecorator[ActivityT], BasicHandler[ActivityT]]
InvokeHandlerUnion = Union[InvokeHandlerDecorator[ActivityT, ResponseT], InvokeHandler[ActivityT, ResponseT]]
VoidInvokeHandlerUnion = Union[VoidInvokeHandlerDecorator[ActivityT], VoidInvokeHandler[ActivityT]]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass, field
from logging import Logger
from typing import Any, Awaitable, Callable, List, Optional, TypedDict, Union, cast

from microsoft.teams.api import ApiClientSettings
from microsoft.teams.common import Storage
from typing_extensions import Unpack

from .plugins import PluginBase


class AppOptions(TypedDict, total=False):
    """Configuration options for the Teams App."""

    client_id: Optional[str]
    """The client ID of the app registration."""
    client_secret: Optional[str]
    """The client secret. If provided with client_id, uses ClientCredentials auth."""
    tenant_id: Optional[str]
    """The tenant ID. Required for single-tenant apps."""
    # Custom token provider function
    token: Optional[Callable[[Union[str, list[str]], Optional[str]], Union[str, Awaitable[str]]]]
    """Custom token provider function. If provided with client_id (no client_secret), uses TokenCredentials."""

    # Managed identity configuration (used when client_id provided without client_secret or token)
    managed_identity_client_id: Optional[str]
    """
    The managed identity client ID for user-assigned managed identity.
    Set to "system" for system-assigned managed identity (triggers Federated Identity Credentials).
    If set to a different client ID than client_id, triggers Federated Identity Credentials with user-assigned MI.
    If not set or equals client_id, uses direct managed identity (no federation).
    """

    # Infrastructure
    logger: Optional[Logger]
    storage: Optional[Storage[str, Any]]
    plugins: Optional[List[PluginBase]]
    skip_auth: Optional[bool]

    # OAuth
    default_connection_name: Optional[str]
    """The OAuth connection name to use for authentication. Defaults to 'graph'."""

    # API Client Settings
    api_client_settings: Optional[ApiClientSettings]
    """API client settings used for overriding."""


@dataclass
class InternalAppOptions:
    """Internal dataclass for AppOptions with defaults and non-nullable fields."""

    # Fields with defaults
    skip_auth: bool = False
    default_connection_name: str = "graph"
    """The OAuth connection name to use for authentication."""
    plugins: List[PluginBase] = field(default_factory=lambda: [])
    api_client_settings: Optional[ApiClientSettings] = None
    """API client settings used for overriding."""

    # Optional fields
    client_id: Optional[str] = None
    """The client ID of the app registration."""
    client_secret: Optional[str] = None
    """The client secret. If provided with client_id, uses ClientCredentials auth."""
    tenant_id: Optional[str] = None
    """The tenant ID. Required for single-tenant apps."""
    token: Optional[Callable[[Union[str, list[str]], Optional[str]], Union[str, Awaitable[str]]]] = None
    """Custom token provider function. If provided with client_id (no client_secret), uses TokenCredentials."""
    managed_identity_client_id: Optional[str] = None
    """
    The managed identity client ID for user-assigned managed identity.
    Set to "system" for system-assigned managed identity (triggers Federated Identity Credentials).
    If set to a different client ID than client_id, triggers Federated Identity Credentials with user-assigned MI.
    If not set or equals client_id, uses direct managed identity (no federation).
    """
    logger: Optional[Logger] = None
    storage: Optional[Storage[str, Any]] = None

    @classmethod
    def from_typeddict(cls, options: AppOptions) -> "InternalAppOptions":
        """
        Create InternalAppOptions from AppOptions TypedDict with defaults applied.

        Args:
            options: AppOptions TypedDict (potentially with None values)

        Returns:
            InternalAppOptions with proper defaults and non-nullable required fields
        """
        kwargs: dict[str, Any] = {k: v for k, v in options.items() if v is not None}
        return cls(**kwargs)


def merge_app_options_with_defaults(**options: Unpack[AppOptions]) -> AppOptions:
    """
    Create AppOptions with default values merged with provided options.

    Args:
        **options: Configuration options to override defaults

    Returns:
        AppOptions with defaults applied
    """
    defaults: AppOptions = {
        "skip_auth": False,
        "default_connection_name": "graph",
        "plugins": [],
    }

    return cast(AppOptions, {**defaults, **options})



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""
# pyright: basic

from unittest.mock import AsyncMock, MagicMock

import pytest
from microsoft.teams.api import Activity, ConversationReference, TokenProtocol
from microsoft.teams.apps import ActivityEvent, ActivityResponseEvent, ActivitySentEvent, ErrorEvent, HttpPlugin, Sender
from microsoft.teams.apps.app_events import EventManager
from microsoft.teams.apps.app_process import ActivityProcessor
from microsoft.teams.common.events.event_emitter import EventEmitter


class TestEventManager:
    """Test cases for the EventManager class."""

    @pytest.fixture
    def mock_event_emitter(self):
        """Create a mock EventEmitter."""
        return MagicMock(spec=EventEmitter)

    @pytest.fixture
    def mock_activity_processor(self):
        """Create a mock ActivityProcessor."""
        return MagicMock(spec=ActivityProcessor)

    @pytest.fixture
    def event_manager(self, mock_event_emitter, mock_activity_processor):
        """Create an EventManager instance."""
        return EventManager(mock_event_emitter, mock_activity_processor)

    @pytest.fixture
    def mock_plugins(self):
        plugin = MagicMock(spec=HttpPlugin)
        plugin.on_error_event = AsyncMock()
        plugin.on_error = AsyncMock()
        plugin.on_activity_sent = AsyncMock()
        plugin.on_activity_response = AsyncMock()
        plugin_two = MagicMock(spec=HttpPlugin)
        return [plugin, plugin_two]

    @pytest.mark.asyncio
    async def test_on_error(self, event_manager, mock_event_emitter, mock_plugins):
        error_event = ErrorEvent(error=Exception("Test Error"))

        await event_manager.on_error(error_event, mock_plugins)

        for plugin in mock_plugins:
            if hasattr(plugin, "on_error_event"):
                plugin.on_error.assert_called()
        mock_event_emitter.emit.assert_called_once_with("error", error_event)

    @pytest.mark.asyncio
    async def test_on_activity(self, event_manager, mock_event_emitter, mock_activity_processor, mock_plugins):
        """Test the on_activity method."""
        activity_event = ActivityEvent(
            sender=Sender(), activity=MagicMock(spec=Activity), token=MagicMock(spec=TokenProtocol)
        )

        await event_manager.on_activity(activity_event, mock_plugins)

        mock_event_emitter.emit.assert_called_once_with("activity", activity_event)
        mock_activity_processor.process_activity.assert_called_once_with(
            plugins=mock_plugins, sender=activity_event.sender, event=activity_event
        )

    @pytest.mark.asyncio
    async def test_on_activity_sent(self, event_manager, mock_event_emitter, mock_plugins):
        """Test the on_activity_sent method."""
        sender = Sender()
        activity_sent_event = ActivitySentEvent(
            sender=sender, activity=MagicMock(spec=Activity), conversation_ref=MagicMock(spec=ConversationReference)
        )

        await event_manager.on_activity_sent(sender, activity_sent_event, mock_plugins)

        for plugin in mock_plugins:
            if callable(plugin.on_activity_sent):
                plugin.on_activity_sent.assert_called()
        mock_event_emitter.emit.assert_called_once_with(
            "activity_sent", {"event": activity_sent_event, "sender": sender}
        )

    @pytest.mark.asyncio
    async def test_on_activity_response(self, event_manager, mock_event_emitter, mock_plugins):
        """Test the on_activity_response method."""
        sender = Sender()
        activity_response_event = ActivityResponseEvent(
            activity=MagicMock(spec=Activity), response=MagicMock(), conversation_ref=MagicMock()
        )

        await event_manager.on_activity_response(sender, activity_response_event, mock_plugins)

        for plugin in mock_plugins:
            if callable(plugin.on_activity_response):
                plugin.on_activity_response.assert_called()
        mock_event_emitter.emit.assert_called_once_with(
            "activity_response", {"event": activity_response_event, "sender": sender}
        )



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from typing import Literal, cast
from unittest.mock import MagicMock, create_autospec, patch

import pytest
from microsoft.teams.api import (
    ClientCredentials,
    FederatedIdentityCredentials,
    JsonWebToken,
    ManagedIdentityCredentials,
)
from microsoft.teams.apps.token_manager import TokenManager
from msal import ManagedIdentityClient  # pyright: ignore[reportMissingTypeStubs]

# Valid JWT-like token for testing (format: header.payload.signature)
VALID_TEST_TOKEN = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9."
    "eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiaWF0IjoxNTE2MjM5MDIyfQ."
    "SflKxwRJSMeKKF2QT4fwpMeJf36POk6yJV_adQssw5c"
)


class TestTokenManager:
    """Test TokenManager functionality."""

    @pytest.mark.asyncio
    async def test_get_bot_token_success(self):
        """Test successful bot token retrieval using MSAL."""
        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="test-tenant-id",
        )

        # Mock MSAL ConfidentialClientApplication
        mock_msal_app = MagicMock()
        mock_msal_app.acquire_token_for_client = MagicMock(return_value={"access_token": VALID_TEST_TOKEN})

        with patch(
            "microsoft.teams.apps.token_manager.ConfidentialClientApplication", return_value=mock_msal_app
        ) as mock_msal_class:
            manager = TokenManager(credentials=mock_credentials)

            token = await manager.get_bot_token()

            assert token is not None
            assert isinstance(token, JsonWebToken)
            assert str(token) == VALID_TEST_TOKEN

            # Verify MSAL was called with correct scope
            mock_msal_app.acquire_token_for_client.assert_called_once_with(["https://api.botframework.com/.default"])

            # Verify MSAL app was created with the credentials tenant_id
            mock_msal_class.assert_called_once_with(
                "test-client-id",
                client_credential="test-client-secret",
                authority="https://login.microsoftonline.com/test-tenant-id",
            )

    @pytest.mark.asyncio
    async def test_get_bot_token_no_credentials(self):
        """Test getting bot token with no credentials returns None."""
        manager = TokenManager(credentials=None)
        token = await manager.get_bot_token()
        assert token is None

    @pytest.mark.asyncio
    async def test_get_bot_token_default_tenant(self):
        """Test bot token uses default tenant when credentials have no tenant_id."""
        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id=None,
        )

        # Mock MSAL ConfidentialClientApplication
        mock_msal_app = MagicMock()
        mock_msal_app.acquire_token_for_client = MagicMock(return_value={"access_token": VALID_TEST_TOKEN})

        with patch(
            "microsoft.teams.apps.token_manager.ConfidentialClientApplication", return_value=mock_msal_app
        ) as mock_msal_class:
            manager = TokenManager(credentials=mock_credentials)

            token = await manager.get_bot_token()

            assert token is not None
            assert isinstance(token, JsonWebToken)

            # Verify MSAL app was created with default bot token tenant
            mock_msal_class.assert_called_once_with(
                "test-client-id",
                client_credential="test-client-secret",
                authority="https://login.microsoftonline.com/botframework.com",
            )

    @pytest.mark.asyncio
    async def test_get_graph_token_default(self):
        """Test getting default graph token using MSAL."""
        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="default-tenant-id",
        )

        # Mock MSAL ConfidentialClientApplication
        mock_msal_app = MagicMock()
        mock_msal_app.acquire_token_for_client = MagicMock(return_value={"access_token": VALID_TEST_TOKEN})

        with patch(
            "microsoft.teams.apps.token_manager.ConfidentialClientApplication", return_value=mock_msal_app
        ) as mock_msal_class:
            manager = TokenManager(credentials=mock_credentials)

            token = await manager.get_graph_token()

            assert token is not None
            assert isinstance(token, JsonWebToken)
            assert str(token) == VALID_TEST_TOKEN

            # Verify MSAL was called with correct scope
            mock_msal_app.acquire_token_for_client.assert_called_once_with(["https://graph.microsoft.com/.default"])

            # Verify MSAL app was created with the credentials tenant_id
            mock_msal_class.assert_called_once_with(
                "test-client-id",
                client_credential="test-client-secret",
                authority="https://login.microsoftonline.com/default-tenant-id",
            )

    @pytest.mark.asyncio
    async def test_get_graph_token_default_tenant(self):
        """Test graph token uses default tenant when credentials have no tenant_id."""
        mock_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id=None,
        )

        # Mock MSAL ConfidentialClientApplication
        mock_msal_app = MagicMock()
        mock_msal_app.acquire_token_for_client = MagicMock(return_value={"access_token": VALID_TEST_TOKEN})

        with patch(
            "microsoft.teams.apps.token_manager.ConfidentialClientApplication", return_value=mock_msal_app
        ) as mock_msal_class:
            manager = TokenManager(credentials=mock_credentials)

            token = await manager.get_graph_token()

            assert token is not None
            assert isinstance(token, JsonWebToken)

            # Verify MSAL app was created with default graph token tenant "common"
            mock_msal_class.assert_called_once_with(
                "test-client-id",
                client_credential="test-client-secret",
                authority="https://login.microsoftonline.com/common",
            )

    @pytest.mark.asyncio
    async def test_get_graph_token_with_tenant(self):
        """Test getting tenant-specific graph token using MSAL."""
        original_credentials = ClientCredentials(
            client_id="test-client-id",
            client_secret="test-client-secret",
            tenant_id="original-tenant-id",
        )

        # Mock MSAL ConfidentialClientApplication
        mock_msal_app = MagicMock()
        mock_msal_app.acquire_token_for_client = MagicMock(return_value={"access_token": VALID_TEST_TOKEN})

        with patch(
            "microsoft.teams.apps.token_manager.ConfidentialClientApplication", return_value=mock_msal_app
        ) as mock_msal_class:
            manager = TokenManager(credentials=original_credentials)

            token = await manager.get_graph_token("different-tenant-id")

            assert token is not None
            assert isinstance(token, JsonWebToken)

            # Verify MSAL app was created with different tenant ID
            # The manager caches MSAL clients, so we check the call to the class constructor
            calls = mock_msal_class.call_args_list
            # Should have been called with different-tenant-id
            assert any("different-tenant-id" in str(call) for call in calls)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "get_token_method,expected_resource",
        [
            ("get_bot_token", "https://api.botframework.com"),
            ("get_graph_token", "https://graph.microsoft.com"),
        ],
    )
    async def test_get_token_with_managed_identity(self, get_token_method: str, expected_resource: str):
        """Test token retrieval using ManagedIdentityCredentials."""
        mock_credentials = ManagedIdentityCredentials(
            client_id="test-managed-identity-client-id",
            tenant_id="test-tenant-id",
        )

        # Create a mock that will pass isinstance checks
        mock_msal_client = create_autospec(ManagedIdentityClient, instance=True)
        mock_msal_client.acquire_token_for_client.return_value = {"access_token": VALID_TEST_TOKEN}

        manager = TokenManager(credentials=mock_credentials)

        # Patch _get_managed_identity_client to return our mock
        with patch.object(manager, "_get_managed_identity_client", return_value=mock_msal_client):
            # Call the method dynamically
            token = await getattr(manager, get_token_method)()

            assert token is not None
            assert isinstance(token, JsonWebToken)
            assert str(token) == VALID_TEST_TOKEN

            # Verify MSAL was called with resource parameter (not scopes list)
            # and without /.default suffix
            mock_msal_client.acquire_token_for_client.assert_called_once_with(resource=expected_resource)

    @pytest.mark.asyncio
    async def test_get_token_error_handling_with_managed_identity(self):
        """Test error handling when token acquisition fails with ManagedIdentityCredentials."""
        mock_credentials = ManagedIdentityCredentials(
            client_id="test-managed-identity-client-id",
            tenant_id="test-tenant-id",
        )

        # Create a mock that returns an error
        mock_msal_client = create_autospec(ManagedIdentityClient, instance=True)
        mock_msal_client.acquire_token_for_client.return_value = {
            "error": "invalid_client",
            "error_description": "Invalid managed identity configuration",
        }

        manager = TokenManager(credentials=mock_credentials)

        # Patch _get_managed_identity_client to return our mock
        with patch.object(manager, "_get_managed_identity_client", return_value=mock_msal_client):
            # Should raise an error when token acquisition fails
            with pytest.raises(ValueError) as exc_info:
                await manager.get_bot_token()

            assert "invalid_client" in str(exc_info.value)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "mi_type,mi_client_id,description",
        [
            ("system", None, "system-assigned managed identity"),
            ("user", "test-user-mi-client-id", "user-assigned managed identity"),
        ],
    )
    async def test_get_token_with_federated_identity(self, mi_type: str, mi_client_id: str | None, description: str):
        """Test token retrieval using FederatedIdentityCredentials (two-step flow)."""
        mock_credentials = FederatedIdentityCredentials(
            client_id="test-app-client-id",
            managed_identity_type=cast(Literal["system", "user"], mi_type),
            managed_identity_client_id=mi_client_id,
            tenant_id="test-tenant-id",
        )

        manager = TokenManager(credentials=mock_credentials)

        # Mock the managed identity token acquisition (step 1)
        mi_token = "mi_token_from_step_1"
        with patch.object(manager, "_acquire_managed_identity_token", return_value=mi_token):
            # Mock ConfidentialClientApplication for step 2
            with patch("microsoft.teams.apps.token_manager.ConfidentialClientApplication") as mock_confidential_app:
                mock_app_instance = MagicMock()
                mock_app_instance.acquire_token_for_client.return_value = {"access_token": VALID_TEST_TOKEN}
                mock_confidential_app.return_value = mock_app_instance

                token = await manager.get_bot_token()

                assert token is not None, f"Failed for: {description}"
                assert isinstance(token, JsonWebToken), f"Failed for: {description}"
                assert str(token) == VALID_TEST_TOKEN, f"Failed for: {description}"

                # Verify ConfidentialClientApplication was called with MI token as client_assertion
                mock_confidential_app.assert_called_once()
                call_kwargs = mock_confidential_app.call_args[1]
                assert call_kwargs["client_credential"] == {"client_assertion": mi_token}, f"Failed for: {description}"

    @pytest.mark.asyncio
    async def test_get_token_with_federated_identity_step1_failure(self):
        """Test error handling when step 1 (MI token acquisition) fails."""
        mock_credentials = FederatedIdentityCredentials(
            client_id="test-app-client-id",
            managed_identity_type="user",
            managed_identity_client_id="test-mi-client-id",
            tenant_id="test-tenant-id",
        )

        manager = TokenManager(credentials=mock_credentials)

        # Mock step 1 to fail
        with patch.object(
            manager, "_acquire_managed_identity_token", side_effect=ValueError("MI token acquisition failed")
        ):
            with pytest.raises(ValueError) as exc_info:
                await manager.get_bot_token()

            assert "MI token acquisition failed" in str(exc_info.value)

    @pytest.mark.asyncio
    async def test_get_token_with_federated_identity_step2_failure(self):
        """Test error handling when step 2 (final token acquisition) fails."""
        mock_credentials = FederatedIdentityCredentials(
            client_id="test-app-client-id",
            managed_identity_type="user",
            managed_identity_client_id="test-mi-client-id",
            tenant_id="test-tenant-id",
        )

        manager = TokenManager(credentials=mock_credentials)

        # Mock step 1 to succeed
        mi_token = "mi_token_from_step_1"
        with patch.object(manager, "_acquire_managed_identity_token", return_value=mi_token):
            # Mock step 2 to fail
            with patch("microsoft.teams.apps.token_manager.ConfidentialClientApplication") as mock_confidential_app:
                mock_app_instance = MagicMock()
                mock_app_instance.acquire_token_for_client.return_value = {
                    "error": "invalid_grant",
                    "error_description": "FIC Step 2 failed",
                }
                mock_confidential_app.return_value = mock_app_instance

                with pytest.raises(ValueError) as exc_info:
                    await manager.get_bot_token()

                assert "invalid_grant" in str(exc_info.value)



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

import asyncio
import logging
from inspect import isawaitable
from typing import Any, Optional

import requests
from microsoft.teams.api import (
    ClientCredentials,
    Credentials,
    JsonWebToken,
    TokenProtocol,
)
from microsoft.teams.api.auth.credentials import (
    FederatedIdentityCredentials,
    ManagedIdentityCredentials,
    TokenCredentials,
)
from microsoft.teams.common import ConsoleLogger
from msal import (
    ConfidentialClientApplication,
    ManagedIdentityClient,
    SystemAssignedManagedIdentity,
    UserAssignedManagedIdentity,
)

BOT_TOKEN_SCOPE = "https://api.botframework.com/.default"
GRAPH_TOKEN_SCOPE = "https://graph.microsoft.com/.default"
DEFAULT_TENANT_FOR_BOT_TOKEN = "botframework.com"
DEFAULT_TENANT_FOR_GRAPH_TOKEN = "common"
DEFAULT_TOKEN_AUTHORITY = "https://login.microsoftonline.com/{tenant_id}"


class TokenManager:
    """Manages authentication tokens for the Teams application."""

    def __init__(
        self,
        credentials: Optional[Credentials],
        logger: Optional[logging.Logger] = None,
    ):
        self._credentials = credentials

        if not logger:
            self._logger = ConsoleLogger().create_logger("TokenManager")
        else:
            self._logger = logger.getChild("TokenManager")

        self._confidential_clients_by_tenant: dict[str, ConfidentialClientApplication] = {}
        self._managed_identity_client: Optional[ManagedIdentityClient] = None

    async def get_bot_token(self) -> Optional[TokenProtocol]:
        """Refresh the bot authentication token."""
        return await self._get_token(
            BOT_TOKEN_SCOPE, tenant_id=self._resolve_tenant_id(None, DEFAULT_TENANT_FOR_BOT_TOKEN)
        )

    async def get_graph_token(self, tenant_id: Optional[str] = None) -> Optional[TokenProtocol]:
        """
        Get or refresh a Graph API token.

        Args:
            tenant_id: If provided, gets a tenant-specific token. Otherwise uses app's default.
            force: Force refresh even if token is not expired

        Returns:
            The graph token or None if not available
        """
        return await self._get_token(
            GRAPH_TOKEN_SCOPE, tenant_id=self._resolve_tenant_id(tenant_id, DEFAULT_TENANT_FOR_GRAPH_TOKEN)
        )

    async def _get_token(
        self, scope: str, tenant_id: str, *, caller_name: str | None = None
    ) -> Optional[TokenProtocol]:
        credentials = self._credentials
        if self._credentials is None:
            if caller_name:
                self._logger.debug(f"No credentials provided for {caller_name}")
            return None
        if isinstance(credentials, ClientCredentials):
            return await self._get_token_with_client_credentials(credentials, scope, tenant_id)
        elif isinstance(credentials, ManagedIdentityCredentials):
            return await self._get_token_with_managed_identity(credentials, scope)
        elif isinstance(credentials, FederatedIdentityCredentials):
            return await self._get_token_with_federated_identity(credentials, scope, tenant_id)
        elif isinstance(credentials, TokenCredentials):
            return await self._get_token_with_token_provider(credentials, scope, tenant_id)

        return None

    async def _get_token_with_client_credentials(
        self,
        credentials: ClientCredentials,
        scope: str,
        tenant_id: str,
    ) -> TokenProtocol:
        """Get token using ClientCredentials (client secret)."""
        confidential_client = self._get_confidential_client(credentials, tenant_id)

        # ConfidentialClientApplication expects scopes as a list
        token_res: dict[str, Any] = await asyncio.to_thread(
            lambda: confidential_client.acquire_token_for_client([scope])
        )

        return self._handle_token_response(token_res)

    async def _get_token_with_managed_identity(
        self,
        credentials: ManagedIdentityCredentials,
        scope: str,
    ) -> TokenProtocol:
        """Get token using ManagedIdentityCredentials (direct, no federation)."""
        mi_client = self._get_managed_identity_client(credentials)

        # ManagedIdentityClient expects resource as a keyword-only string parameter
        resource = scope.removesuffix("/.default")
        token_res: dict[str, Any] = await asyncio.to_thread(
            lambda: mi_client.acquire_token_for_client(resource=resource)
        )

        return self._handle_token_response(token_res)

    async def _get_token_with_federated_identity(
        self,
        credentials: FederatedIdentityCredentials,
        scope: str,
        tenant_id: str,
    ) -> TokenProtocol:
        """Get token using Federated Identity Credentials (two-step flow)."""

        # Step 1: Get MI token from api://AzureADTokenExchange
        mi_token = await self._acquire_managed_identity_token(credentials)

        # Step 2: Use MI token as client_assertion to get final access token
        confidential_client = ConfidentialClientApplication(
            credentials.client_id,
            client_credential={"client_assertion": mi_token},
            authority=DEFAULT_TOKEN_AUTHORITY.format(tenant_id=tenant_id),
        )

        token_res: dict[str, Any] = await asyncio.to_thread(
            lambda: confidential_client.acquire_token_for_client([scope])
        )

        return self._handle_token_response(token_res, error_prefix="FIC Step 2 failed")

    async def _acquire_managed_identity_token(self, credentials: FederatedIdentityCredentials) -> str:
        """Acquire managed identity token for federated identity credentials."""
        # Use shared method to get or create the managed identity client
        mi_client = self._get_managed_identity_client(credentials)

        mi_token_res: dict[str, Any] = await asyncio.to_thread(
            lambda: mi_client.acquire_token_for_client(resource="api://AzureADTokenExchange")
        )

        if not mi_token_res.get("access_token"):
            self._logger.error("FIC Step 1 failed: Could not acquire MI token")
            error = mi_token_res.get("error", ValueError("Error retrieving MI token"))
            if not isinstance(error, BaseException):
                error = ValueError(error)
            raise error

        return mi_token_res["access_token"]

    async def _get_token_with_token_provider(
        self,
        credentials: TokenCredentials,
        scope: str,
        tenant_id: str,
    ) -> TokenProtocol:
        """Get token using custom token provider function."""
        token = credentials.token(scope, tenant_id)

        if isawaitable(token):
            access_token = await token
        else:
            access_token = token

        return JsonWebToken(access_token)

    def _handle_token_response(self, token_res: dict[str, Any], error_prefix: str = "") -> TokenProtocol:
        """Handle token response from MSAL client."""
        if token_res.get("access_token", None):
            access_token = token_res["access_token"]
            return JsonWebToken(access_token)
        else:
            error_msg = f"{error_prefix}: " if error_prefix else ""
            self._logger.error(f"{error_msg}Could not acquire access token")
            self._logger.debug(f"TokenRes: {token_res}")

            error = token_res.get("error", "Error retrieving token")
            if not isinstance(error, BaseException):
                error = ValueError(error)

            error_description = token_res.get("error_description", "Error retrieving token from MSAL")
            self._logger.error(error_description)
            raise error

    def _get_confidential_client(self, credentials: ClientCredentials, tenant_id: str) -> ConfidentialClientApplication:
        """Get or create ConfidentialClientApplication for ClientCredentials."""
        # Check if client already exists in cache
        cached_client = self._confidential_clients_by_tenant.get(tenant_id)
        if cached_client:
            return cached_client

        client: ConfidentialClientApplication = ConfidentialClientApplication(
            credentials.client_id,
            client_credential=credentials.client_secret,
            authority=f"https://login.microsoftonline.com/{tenant_id}",
        )
        self._confidential_clients_by_tenant[tenant_id] = client
        return client

    def _get_managed_identity_client(
        self, credentials: ManagedIdentityCredentials | FederatedIdentityCredentials
    ) -> ManagedIdentityClient:
        """Get or create ManagedIdentityClient for ManagedIdentityCredentials or FederatedIdentityCredentials."""
        # Check if client already exists in cache

        # ManagedIdentityClient is tenant-agnostic, cache single instance
        if self._managed_identity_client:
            return self._managed_identity_client

        # Determine managed identity type
        if isinstance(credentials, FederatedIdentityCredentials):
            if credentials.managed_identity_type == "system":
                managed_identity = SystemAssignedManagedIdentity()
            else:  # "user"
                mi_client_id = credentials.managed_identity_client_id or credentials.client_id
                managed_identity = UserAssignedManagedIdentity(client_id=mi_client_id)
        else:  # ManagedIdentityCredentials
            # ManagedIdentityCredentials only supports user-assigned
            managed_identity = UserAssignedManagedIdentity(client_id=credentials.client_id)

        self._managed_identity_client = ManagedIdentityClient(
            managed_identity,
            http_client=requests.Session(),
        )
        return self._managed_identity_client

    def _resolve_tenant_id(self, tenant_id: str | None, default_tenant_id: str):
        return tenant_id or (self._credentials.tenant_id if self._credentials else False) or default_tenant_id



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from dataclasses import dataclass
from logging import Logger
from typing import Any, Dict, List, Optional

import jwt
from microsoft.teams.common.logging import ConsoleLogger

JWT_LEEWAY_SECONDS = 300  # Allowable clock skew when validating JWTs


@dataclass
class JwtValidationOptions:
    """Configuration for JWT validation."""

    valid_issuers: List[str]
    """ List of valid issuers for the JWT"""
    valid_audiences: List[str]
    """ List of valid audiences for the JWT"""
    jwks_uri: str
    """ URI to the JSON Web Key Set (JWKS) for token signature verification """
    service_url: Optional[str] = None
    """ Optional service URL to validate against token claims """
    scope: Optional[str] = None
    """ Optional scope that must be present in the token """
    clock_tolerance: int = JWT_LEEWAY_SECONDS
    """ Allowable clock skew when validating JWTs """


class TokenValidator:
    """
    JWT token validator using PyJWKClient for simplified validation.
    """

    def __init__(self, jwt_validation_options: JwtValidationOptions, logger: Optional[Logger] = None):
        """
        Initialize the token validator.

        Args:
            jwt_validation_options: Configuration for JWT validation
        """
        self.logger = logger or ConsoleLogger().create_logger("@teams/token-validator")
        self.options = jwt_validation_options

    # ----- Factory constructors -----
    @classmethod
    def for_service(
        cls, app_id: str, logger: Optional[Logger] = None, service_url: Optional[str] = None
    ) -> "TokenValidator":
        """Create a validator for Bot Framework service tokens.

        Reference: https://learn.microsoft.com/en-us/azure/bot-service/rest-api/bot-framework-rest-connector-authentication

        Args:
            app_id: The bot's Microsoft App ID (used for audience validation)
            service_url: Optional service URL to validate against token claims
            logger: Optional logger instance"""

        options = JwtValidationOptions(
            valid_issuers=["https://api.botframework.com"],
            valid_audiences=[app_id, f"api://{app_id}"],
            jwks_uri="https://login.botframework.com/v1/.well-known/keys",
            service_url=service_url,
        )
        return cls(options, logger)

    @classmethod
    def for_entra(
        cls, app_id: str, tenant_id: Optional[str], scope: Optional[str] = None, logger: Optional[Logger] = None
    ) -> "TokenValidator":
        """Create a validator for Entra ID tokens.

        Args:
            app_id: The app's Microsoft App ID (used for audience validation)
            tenant_id: The Azure AD tenant ID
            scope: Optional scope that must be present in the token
            logger: Optional logger instance

        """

        valid_issuers: List[str] = []
        if tenant_id:
            valid_issuers.append(f"https://login.microsoftonline.com/{tenant_id}/v2.0")
        tenant_id = tenant_id or "common"
        options = JwtValidationOptions(
            valid_issuers=valid_issuers,
            valid_audiences=[app_id, f"api://{app_id}"],
            jwks_uri=f"https://login.microsoftonline.com/{tenant_id}/discovery/v2.0/keys",
            scope=scope,
        )
        return cls(options, logger)

    async def validate_token(
        self, raw_token: str, service_url: Optional[str] = None, scope: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Validate a JWT token.

        Args:
            raw_token: The raw JWT token string
            service_url: Optional service URL to validate against token claims
            scope: Optional scope that must be present in the token

        Returns:
            The decoded JWT payload if validation is successful

        Raises:
            jwt.InvalidTokenError: When token validation fails
        """
        if not raw_token:
            self.logger.error("No token provided")
            raise jwt.InvalidTokenError("No token provided")

        try:
            jwks_client = jwt.PyJWKClient(self.options.jwks_uri)
            # Get signing key automatically from JWKS
            signing_key = jwks_client.get_signing_key_from_jwt(raw_token)

            # Validate token
            payload: Dict[str, Any] = jwt.decode(
                raw_token,
                signing_key.key,
                algorithms=["RS256"],
                audience=self.options.valid_audiences,
                issuer=self.options.valid_issuers,
                options={
                    "verify_signature": True,
                    "verify_aud": True,
                    "verify_iss": bool(self.options.valid_issuers),
                    "verify_exp": True,
                    "verify_iat": True,
                },
                leeway=JWT_LEEWAY_SECONDS,
            )

            # Optional service URL validation
            expected_service_url = service_url or self.options.service_url
            if expected_service_url:
                self._validate_service_url(payload, expected_service_url)

            required_scope = scope or self.options.scope
            if required_scope:
                self._validate_scope(payload, required_scope)

            self.logger.debug("Token validation successful")
            return payload

        except jwt.InvalidTokenError as e:
            self.logger.error(f"Token validation failed: {e}")
            raise

    def _validate_service_url(self, payload: Dict[str, Any], expected_service_url: str) -> None:
        """Validate service URL claim matches expected service URL.

        Args:
            payload: The decoded JWT payload
            expected_service_url: The service URL to validate against
        """
        token_service_url = payload.get("serviceurl")

        if not token_service_url:
            self.logger.error("Token missing serviceurl claim")
            raise jwt.InvalidTokenError("Token missing serviceurl claim")

        # Normalize URLs (remove trailing slashes)
        normalized_token_url = token_service_url.rstrip("/")
        normalized_expected_url = expected_service_url.rstrip("/")

        if normalized_token_url != normalized_expected_url:
            self.logger.error(
                f"Service URL mismatch. Token: {normalized_token_url}, Expected: {normalized_expected_url}"
            )
            raise jwt.InvalidTokenError(
                f"Service URL mismatch. Token: {normalized_token_url}, Expected: {normalized_expected_url}"
            )

    def _validate_scope(self, payload: Dict[str, Any], required_scope: str) -> None:
        """Validate that the required scope is present in the token.

        Args:
            payload: The decoded JWT payload
            required_scope: The scope required to be present in the token
        """
        scopes = payload.get("scp", "") or ""
        if required_scope not in scopes:
            self.logger.error(f"Token missing required scope: {required_scope}")
            raise jwt.InvalidTokenError(f"Token missing required scope: {required_scope}")



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .client_context import ClientContext
from .function_context import FunctionContext

__all__ = ["ClientContext", "FunctionContext"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""
# pyright: basic

from logging import Logger
from typing import Annotated, Callable
from unittest.mock import MagicMock

import pytest
from microsoft.teams.apps import (
    DependencyMetadata,
    ErrorEvent,
    EventMetadata,
    LoggerDependencyOptions,
    Plugin,
    PluginBase,
)
from microsoft.teams.apps.app_events import EventManager
from microsoft.teams.apps.app_plugins import PluginProcessor
from microsoft.teams.apps.container import Container
from microsoft.teams.common import Client, EventEmitter


class TestPluginProcessor:
    """Test cases for PluginProcessor class."""

    @pytest.fixture
    def mock_logger(self):
        return MagicMock(spec=Logger)

    @pytest.fixture
    def container(self):
        return Container()

    @pytest.fixture
    def mock_event_manager(self):
        return MagicMock(spec=EventManager)

    @pytest.fixture
    def mock_client(self):
        return MagicMock(spec=Client)

    @pytest.fixture
    def mock_event_emitter(self):
        return MagicMock(spec=EventEmitter)

    @pytest.fixture
    def plugin_processor(self, container, mock_event_emitter, mock_logger, mock_event_manager):
        """Create a PluginProcessor instance."""
        return PluginProcessor(
            container=container,
            event_manager=mock_event_manager,
            event_emitter=mock_event_emitter,
            logger=mock_logger,
        )

    @pytest.fixture
    def mock_plugin(self):
        """Create a test mock plugin."""

        @Plugin(name="MockPlugin", version="1.0", description="A mock plugin for testing")
        class MockPlugin(PluginBase):
            logger: Annotated[Logger, LoggerDependencyOptions()]
            on_error_event: Annotated[Callable[[ErrorEvent], None], EventMetadata(name="error")]
            client: Annotated[Client, DependencyMetadata()]

            def __init__(self):
                super().__init__()

        return MockPlugin()

    @pytest.mark.asyncio
    async def test_initialize_plugins(self, plugin_processor, mock_plugin):
        """Test the initialize_plugins method."""

        initialized_plugins = plugin_processor.initialize_plugins([mock_plugin])

        assert mock_plugin in initialized_plugins
        assert plugin_processor.get_plugin("MockPlugin") == mock_plugin
        assert plugin_processor.container.MockPlugin() == mock_plugin

    @pytest.mark.asyncio
    async def test_duplicate_in_initialize_plugins(self, plugin_processor, mock_plugin):
        """Test the initialize_plugins method."""

        plugin_processor.initialize_plugins([mock_plugin])

        with pytest.raises(ValueError, match="duplicate plugin MockPlugin found"):
            plugin_processor.initialize_plugins([mock_plugin])

    @pytest.mark.asyncio
    async def test_get_plugin(self, plugin_processor, mock_plugin):
        """Test the get_plugin method."""

        plugin_processor.initialize_plugins([mock_plugin])

        plugin = plugin_processor.get_plugin("MockPlugin")
        assert plugin == mock_plugin

        plugin_none = plugin_processor.get_plugin("NonExistentPlugin")
        assert plugin_none is None

    @pytest.mark.asyncio
    async def test_inject(self, plugin_processor, mock_plugin, mock_logger, mock_client):
        """Test the inject method."""

        plugin_processor.container.set_provider("logger", mock_logger)
        plugin_processor.container.set_provider("client", mock_client)

        plugin_processor.initialize_plugins([mock_plugin])
        plugin_processor.inject(mock_plugin)

        assert mock_plugin.logger._extract_mock_name() == "mock().getChild()"
        assert hasattr(mock_plugin, "on_error_event")
        assert mock_plugin.client._extract_mock_name() == "mock()"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from logging import Logger
from typing import Awaitable, Callable

import jwt
from fastapi import HTTPException, Request, Response
from microsoft.teams.api.auth.json_web_token import JsonWebToken

from .token_validator import TokenValidator


def create_jwt_validation_middleware(
    app_id: str,
    logger: Logger,
    paths: list[str],
):
    """
    Create JWT validation middleware instance.

    Args:
        app_id: Bot's Microsoft App ID for audience validation
        logger: Logger instance
        paths: List of paths to validate

    Returns:
        Middleware function that can be added to FastAPI app
    """
    # Create service token validator
    token_validator = TokenValidator.for_service(app_id, logger)

    async def middleware(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        """JWT validation middleware function."""
        # Only validate specified paths
        if request.url.path not in paths:
            return await call_next(request)

        # Extract Bearer token
        authorization = request.headers.get("authorization")
        if not authorization or not authorization.startswith("Bearer "):
            logger.warning("Unauthorized request - missing or invalid authorization header")
            raise HTTPException(status_code=401, detail="unauthorized")

        raw_token = authorization.removeprefix("Bearer ")

        try:
            # Parse request body to get service URL for validation
            body = await request.json()
            service_url = body.get("serviceUrl")

            # Validate token
            await token_validator.validate_token(raw_token, service_url)
            validated_token = JsonWebToken(value=raw_token)

            logger.debug(f"Validated service token for activity {body.get('id', 'unknown')}")

            # Store validated token in request state
            request.state.validated_token = validated_token

            return await call_next(request)

        except jwt.InvalidTokenError as e:
            logger.warning(f"JWT token validation failed: {e}")
            raise HTTPException(status_code=401, detail="unauthorized") from e
        except Exception as e:
            logger.error(f"Unexpected error during token validation: {e}")
            raise HTTPException(status_code=500, detail="internal server error") from e

    return middleware



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.

GENERATED FILE - DO NOT EDIT MANUALLY
This file is generated by generate_handlers.py based on activity_config.py

To regenerate, run:
uv run generate-activity-handlers
"""

from abc import ABC, abstractmethod
from logging import Logger
from typing import Callable, Optional, overload

from microsoft.teams.api.activities import (
    Activity,
    AdaptiveCardInvokeActivity,
    CommandResultActivity,
    CommandSendActivity,
    ConfigFetchInvokeActivity,
    ConfigSubmitInvokeActivity,
    ConversationUpdateActivity,
    EndOfConversationActivity,
    EventActivity,
    ExecuteActionInvokeActivity,
    FileConsentInvokeActivity,
    HandoffActionInvokeActivity,
    HandoffActivity,
    InstalledActivity,
    InstallUpdateActivity,
    InvokeActivity,
    MeetingEndEventActivity,
    MeetingParticipantJoinEventActivity,
    MeetingParticipantLeaveEventActivity,
    MeetingStartEventActivity,
    MessageActivity,
    MessageDeleteActivity,
    MessageExtensionAnonQueryLinkInvokeActivity,
    MessageExtensionCardButtonClickedInvokeActivity,
    MessageExtensionFetchTaskInvokeActivity,
    MessageExtensionQueryInvokeActivity,
    MessageExtensionQueryLinkInvokeActivity,
    MessageExtensionQuerySettingUrlInvokeActivity,
    MessageExtensionSelectItemInvokeActivity,
    MessageExtensionSettingInvokeActivity,
    MessageExtensionSubmitActionInvokeActivity,
    MessageReactionActivity,
    MessageSubmitActionInvokeActivity,
    MessageUpdateActivity,
    ReadReceiptEventActivity,
    SignInTokenExchangeInvokeActivity,
    SignInVerifyStateInvokeActivity,
    TabFetchInvokeActivity,
    TabSubmitInvokeActivity,
    TaskFetchInvokeActivity,
    TaskSubmitInvokeActivity,
    TraceActivity,
    TypingActivity,
    UninstalledActivity,
)
from microsoft.teams.api.models.invoke_response import (
    AdaptiveCardInvokeResponse,
    ConfigInvokeResponse,
    MessagingExtensionActionInvokeResponse,
    MessagingExtensionInvokeResponse,
    TabInvokeResponse,
    TaskModuleInvokeResponse,
    TokenExchangeInvokeResponseType,
)

from .activity_route_configs import ACTIVITY_ROUTES
from .router import ActivityRouter
from .type_helpers import (
    BasicHandler,
    BasicHandlerUnion,
    InvokeHandler,
    InvokeHandlerUnion,
    VoidInvokeHandler,
    VoidInvokeHandlerUnion,
)
from .type_validation import validate_handler_type


class GeneratedActivityHandlerMixin(ABC):
    """Mixin class providing typed activity handler registration methods."""

    @property
    @abstractmethod
    def router(self) -> ActivityRouter:
        """The activity router instance. Must be implemented by the concrete class."""
        pass

    @property
    @abstractmethod
    def logger(self) -> Logger:
        """The logger instance used by the app."""
        pass

    @overload
    def on_message(self, handler: BasicHandler[MessageActivity]) -> BasicHandler[MessageActivity]: ...

    @overload
    def on_message(self) -> Callable[[BasicHandler[MessageActivity]], BasicHandler[MessageActivity]]: ...

    def on_message(self, handler: Optional[BasicHandler[MessageActivity]] = None) -> BasicHandlerUnion[MessageActivity]:
        """Register a message activity handler."""

        def decorator(func: BasicHandler[MessageActivity]) -> BasicHandler[MessageActivity]:
            validate_handler_type(self.logger, func, MessageActivity, "on_message", "MessageActivity")
            config = ACTIVITY_ROUTES["message"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_delete(
        self, handler: BasicHandler[MessageDeleteActivity]
    ) -> BasicHandler[MessageDeleteActivity]: ...

    @overload
    def on_message_delete(
        self,
    ) -> Callable[[BasicHandler[MessageDeleteActivity]], BasicHandler[MessageDeleteActivity]]: ...

    def on_message_delete(
        self, handler: Optional[BasicHandler[MessageDeleteActivity]] = None
    ) -> BasicHandlerUnion[MessageDeleteActivity]:
        """Register a message_delete activity handler."""

        def decorator(func: BasicHandler[MessageDeleteActivity]) -> BasicHandler[MessageDeleteActivity]:
            validate_handler_type(
                self.logger, func, MessageDeleteActivity, "on_message_delete", "MessageDeleteActivity"
            )
            config = ACTIVITY_ROUTES["message_delete"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_soft_delete_message(
        self, handler: BasicHandler[MessageDeleteActivity]
    ) -> BasicHandler[MessageDeleteActivity]: ...

    @overload
    def on_soft_delete_message(
        self,
    ) -> Callable[[BasicHandler[MessageDeleteActivity]], BasicHandler[MessageDeleteActivity]]: ...

    def on_soft_delete_message(
        self, handler: Optional[BasicHandler[MessageDeleteActivity]] = None
    ) -> BasicHandlerUnion[MessageDeleteActivity]:
        """Register a soft_delete_message activity handler."""

        def decorator(func: BasicHandler[MessageDeleteActivity]) -> BasicHandler[MessageDeleteActivity]:
            validate_handler_type(
                self.logger, func, MessageDeleteActivity, "on_soft_delete_message", "MessageDeleteActivity"
            )
            config = ACTIVITY_ROUTES["soft_delete_message"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_reaction(
        self, handler: BasicHandler[MessageReactionActivity]
    ) -> BasicHandler[MessageReactionActivity]: ...

    @overload
    def on_message_reaction(
        self,
    ) -> Callable[[BasicHandler[MessageReactionActivity]], BasicHandler[MessageReactionActivity]]: ...

    def on_message_reaction(
        self, handler: Optional[BasicHandler[MessageReactionActivity]] = None
    ) -> BasicHandlerUnion[MessageReactionActivity]:
        """Register a message_reaction activity handler."""

        def decorator(func: BasicHandler[MessageReactionActivity]) -> BasicHandler[MessageReactionActivity]:
            validate_handler_type(
                self.logger, func, MessageReactionActivity, "on_message_reaction", "MessageReactionActivity"
            )
            config = ACTIVITY_ROUTES["message_reaction"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_update(
        self, handler: BasicHandler[MessageUpdateActivity]
    ) -> BasicHandler[MessageUpdateActivity]: ...

    @overload
    def on_message_update(
        self,
    ) -> Callable[[BasicHandler[MessageUpdateActivity]], BasicHandler[MessageUpdateActivity]]: ...

    def on_message_update(
        self, handler: Optional[BasicHandler[MessageUpdateActivity]] = None
    ) -> BasicHandlerUnion[MessageUpdateActivity]:
        """Register a message_update activity handler."""

        def decorator(func: BasicHandler[MessageUpdateActivity]) -> BasicHandler[MessageUpdateActivity]:
            validate_handler_type(
                self.logger, func, MessageUpdateActivity, "on_message_update", "MessageUpdateActivity"
            )
            config = ACTIVITY_ROUTES["message_update"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_undelete_message(
        self, handler: BasicHandler[MessageUpdateActivity]
    ) -> BasicHandler[MessageUpdateActivity]: ...

    @overload
    def on_undelete_message(
        self,
    ) -> Callable[[BasicHandler[MessageUpdateActivity]], BasicHandler[MessageUpdateActivity]]: ...

    def on_undelete_message(
        self, handler: Optional[BasicHandler[MessageUpdateActivity]] = None
    ) -> BasicHandlerUnion[MessageUpdateActivity]:
        """Register a undelete_message activity handler."""

        def decorator(func: BasicHandler[MessageUpdateActivity]) -> BasicHandler[MessageUpdateActivity]:
            validate_handler_type(
                self.logger, func, MessageUpdateActivity, "on_undelete_message", "MessageUpdateActivity"
            )
            config = ACTIVITY_ROUTES["undelete_message"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_edit_message(self, handler: BasicHandler[MessageUpdateActivity]) -> BasicHandler[MessageUpdateActivity]: ...

    @overload
    def on_edit_message(
        self,
    ) -> Callable[[BasicHandler[MessageUpdateActivity]], BasicHandler[MessageUpdateActivity]]: ...

    def on_edit_message(
        self, handler: Optional[BasicHandler[MessageUpdateActivity]] = None
    ) -> BasicHandlerUnion[MessageUpdateActivity]:
        """Register a edit_message activity handler."""

        def decorator(func: BasicHandler[MessageUpdateActivity]) -> BasicHandler[MessageUpdateActivity]:
            validate_handler_type(self.logger, func, MessageUpdateActivity, "on_edit_message", "MessageUpdateActivity")
            config = ACTIVITY_ROUTES["edit_message"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_command(self, handler: BasicHandler[CommandSendActivity]) -> BasicHandler[CommandSendActivity]: ...

    @overload
    def on_command(self) -> Callable[[BasicHandler[CommandSendActivity]], BasicHandler[CommandSendActivity]]: ...

    def on_command(
        self, handler: Optional[BasicHandler[CommandSendActivity]] = None
    ) -> BasicHandlerUnion[CommandSendActivity]:
        """Register a command activity handler."""

        def decorator(func: BasicHandler[CommandSendActivity]) -> BasicHandler[CommandSendActivity]:
            validate_handler_type(self.logger, func, CommandSendActivity, "on_command", "CommandSendActivity")
            config = ACTIVITY_ROUTES["command"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_command_result(
        self, handler: BasicHandler[CommandResultActivity]
    ) -> BasicHandler[CommandResultActivity]: ...

    @overload
    def on_command_result(
        self,
    ) -> Callable[[BasicHandler[CommandResultActivity]], BasicHandler[CommandResultActivity]]: ...

    def on_command_result(
        self, handler: Optional[BasicHandler[CommandResultActivity]] = None
    ) -> BasicHandlerUnion[CommandResultActivity]:
        """Register a command_result activity handler."""

        def decorator(func: BasicHandler[CommandResultActivity]) -> BasicHandler[CommandResultActivity]:
            validate_handler_type(
                self.logger, func, CommandResultActivity, "on_command_result", "CommandResultActivity"
            )
            config = ACTIVITY_ROUTES["command_result"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_conversation_update(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_conversation_update(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_conversation_update(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a conversation_update activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_conversation_update", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["conversation_update"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_channel_created(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_channel_created(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_channel_created(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a channel_created activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_channel_created", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["channel_created"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_channel_deleted(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_channel_deleted(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_channel_deleted(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a channel_deleted activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_channel_deleted", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["channel_deleted"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_channel_renamed(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_channel_renamed(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_channel_renamed(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a channel_renamed activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_channel_renamed", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["channel_renamed"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_channel_restored(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_channel_restored(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_channel_restored(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a channel_restored activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_channel_restored", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["channel_restored"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_team_archived(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_team_archived(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_team_archived(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a team_archived activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_team_archived", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["team_archived"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_team_deleted(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_team_deleted(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_team_deleted(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a team_deleted activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_team_deleted", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["team_deleted"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_team_hard_deleted(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_team_hard_deleted(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_team_hard_deleted(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a team_hard_deleted activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_team_hard_deleted", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["team_hard_deleted"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_team_renamed(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_team_renamed(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_team_renamed(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a team_renamed activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_team_renamed", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["team_renamed"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_team_restored(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_team_restored(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_team_restored(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a team_restored activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_team_restored", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["team_restored"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_team_unarchived(
        self, handler: BasicHandler[ConversationUpdateActivity]
    ) -> BasicHandler[ConversationUpdateActivity]: ...

    @overload
    def on_team_unarchived(
        self,
    ) -> Callable[[BasicHandler[ConversationUpdateActivity]], BasicHandler[ConversationUpdateActivity]]: ...

    def on_team_unarchived(
        self, handler: Optional[BasicHandler[ConversationUpdateActivity]] = None
    ) -> BasicHandlerUnion[ConversationUpdateActivity]:
        """Register a team_unarchived activity handler."""

        def decorator(func: BasicHandler[ConversationUpdateActivity]) -> BasicHandler[ConversationUpdateActivity]:
            validate_handler_type(
                self.logger, func, ConversationUpdateActivity, "on_team_unarchived", "ConversationUpdateActivity"
            )
            config = ACTIVITY_ROUTES["team_unarchived"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_end_of_conversation(
        self, handler: BasicHandler[EndOfConversationActivity]
    ) -> BasicHandler[EndOfConversationActivity]: ...

    @overload
    def on_end_of_conversation(
        self,
    ) -> Callable[[BasicHandler[EndOfConversationActivity]], BasicHandler[EndOfConversationActivity]]: ...

    def on_end_of_conversation(
        self, handler: Optional[BasicHandler[EndOfConversationActivity]] = None
    ) -> BasicHandlerUnion[EndOfConversationActivity]:
        """Register a end_of_conversation activity handler."""

        def decorator(func: BasicHandler[EndOfConversationActivity]) -> BasicHandler[EndOfConversationActivity]:
            validate_handler_type(
                self.logger, func, EndOfConversationActivity, "on_end_of_conversation", "EndOfConversationActivity"
            )
            config = ACTIVITY_ROUTES["end_of_conversation"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_event(self, handler: BasicHandler[EventActivity]) -> BasicHandler[EventActivity]: ...

    @overload
    def on_event(self) -> Callable[[BasicHandler[EventActivity]], BasicHandler[EventActivity]]: ...

    def on_event(self, handler: Optional[BasicHandler[EventActivity]] = None) -> BasicHandlerUnion[EventActivity]:
        """Register a event activity handler."""

        def decorator(func: BasicHandler[EventActivity]) -> BasicHandler[EventActivity]:
            validate_handler_type(self.logger, func, EventActivity, "on_event", "EventActivity")
            config = ACTIVITY_ROUTES["event"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_read_receipt(
        self, handler: BasicHandler[ReadReceiptEventActivity]
    ) -> BasicHandler[ReadReceiptEventActivity]: ...

    @overload
    def on_read_receipt(
        self,
    ) -> Callable[[BasicHandler[ReadReceiptEventActivity]], BasicHandler[ReadReceiptEventActivity]]: ...

    def on_read_receipt(
        self, handler: Optional[BasicHandler[ReadReceiptEventActivity]] = None
    ) -> BasicHandlerUnion[ReadReceiptEventActivity]:
        """Register a read_receipt activity handler."""

        def decorator(func: BasicHandler[ReadReceiptEventActivity]) -> BasicHandler[ReadReceiptEventActivity]:
            validate_handler_type(
                self.logger, func, ReadReceiptEventActivity, "on_read_receipt", "ReadReceiptEventActivity"
            )
            config = ACTIVITY_ROUTES["read_receipt"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_meeting_start(
        self, handler: BasicHandler[MeetingStartEventActivity]
    ) -> BasicHandler[MeetingStartEventActivity]: ...

    @overload
    def on_meeting_start(
        self,
    ) -> Callable[[BasicHandler[MeetingStartEventActivity]], BasicHandler[MeetingStartEventActivity]]: ...

    def on_meeting_start(
        self, handler: Optional[BasicHandler[MeetingStartEventActivity]] = None
    ) -> BasicHandlerUnion[MeetingStartEventActivity]:
        """Register a meeting_start activity handler."""

        def decorator(func: BasicHandler[MeetingStartEventActivity]) -> BasicHandler[MeetingStartEventActivity]:
            validate_handler_type(
                self.logger, func, MeetingStartEventActivity, "on_meeting_start", "MeetingStartEventActivity"
            )
            config = ACTIVITY_ROUTES["meeting_start"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_meeting_end(
        self, handler: BasicHandler[MeetingEndEventActivity]
    ) -> BasicHandler[MeetingEndEventActivity]: ...

    @overload
    def on_meeting_end(
        self,
    ) -> Callable[[BasicHandler[MeetingEndEventActivity]], BasicHandler[MeetingEndEventActivity]]: ...

    def on_meeting_end(
        self, handler: Optional[BasicHandler[MeetingEndEventActivity]] = None
    ) -> BasicHandlerUnion[MeetingEndEventActivity]:
        """Register a meeting_end activity handler."""

        def decorator(func: BasicHandler[MeetingEndEventActivity]) -> BasicHandler[MeetingEndEventActivity]:
            validate_handler_type(
                self.logger, func, MeetingEndEventActivity, "on_meeting_end", "MeetingEndEventActivity"
            )
            config = ACTIVITY_ROUTES["meeting_end"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_meeting_participant_join(
        self, handler: BasicHandler[MeetingParticipantJoinEventActivity]
    ) -> BasicHandler[MeetingParticipantJoinEventActivity]: ...

    @overload
    def on_meeting_participant_join(
        self,
    ) -> Callable[
        [BasicHandler[MeetingParticipantJoinEventActivity]], BasicHandler[MeetingParticipantJoinEventActivity]
    ]: ...

    def on_meeting_participant_join(
        self, handler: Optional[BasicHandler[MeetingParticipantJoinEventActivity]] = None
    ) -> BasicHandlerUnion[MeetingParticipantJoinEventActivity]:
        """Register a meeting_participant_join activity handler."""

        def decorator(
            func: BasicHandler[MeetingParticipantJoinEventActivity],
        ) -> BasicHandler[MeetingParticipantJoinEventActivity]:
            validate_handler_type(
                self.logger,
                func,
                MeetingParticipantJoinEventActivity,
                "on_meeting_participant_join",
                "MeetingParticipantJoinEventActivity",
            )
            config = ACTIVITY_ROUTES["meeting_participant_join"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_meeting_participant_leave(
        self, handler: BasicHandler[MeetingParticipantLeaveEventActivity]
    ) -> BasicHandler[MeetingParticipantLeaveEventActivity]: ...

    @overload
    def on_meeting_participant_leave(
        self,
    ) -> Callable[
        [BasicHandler[MeetingParticipantLeaveEventActivity]], BasicHandler[MeetingParticipantLeaveEventActivity]
    ]: ...

    def on_meeting_participant_leave(
        self, handler: Optional[BasicHandler[MeetingParticipantLeaveEventActivity]] = None
    ) -> BasicHandlerUnion[MeetingParticipantLeaveEventActivity]:
        """Register a meeting_participant_leave activity handler."""

        def decorator(
            func: BasicHandler[MeetingParticipantLeaveEventActivity],
        ) -> BasicHandler[MeetingParticipantLeaveEventActivity]:
            validate_handler_type(
                self.logger,
                func,
                MeetingParticipantLeaveEventActivity,
                "on_meeting_participant_leave",
                "MeetingParticipantLeaveEventActivity",
            )
            config = ACTIVITY_ROUTES["meeting_participant_leave"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_config_open(
        self, handler: InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse]
    ) -> InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse]: ...

    @overload
    def on_config_open(
        self,
    ) -> Callable[
        [InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse]],
        InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse],
    ]: ...

    def on_config_open(
        self, handler: Optional[InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse]] = None
    ) -> InvokeHandlerUnion[ConfigFetchInvokeActivity, ConfigInvokeResponse]:
        """Register a config.open activity handler."""

        def decorator(
            func: InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse],
        ) -> InvokeHandler[ConfigFetchInvokeActivity, ConfigInvokeResponse]:
            validate_handler_type(
                self.logger, func, ConfigFetchInvokeActivity, "on_config_open", "ConfigFetchInvokeActivity"
            )
            config = ACTIVITY_ROUTES["config.open"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_config_submit(
        self, handler: InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse]
    ) -> InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse]: ...

    @overload
    def on_config_submit(
        self,
    ) -> Callable[
        [InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse]],
        InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse],
    ]: ...

    def on_config_submit(
        self, handler: Optional[InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse]] = None
    ) -> InvokeHandlerUnion[ConfigSubmitInvokeActivity, ConfigInvokeResponse]:
        """Register a config.submit activity handler."""

        def decorator(
            func: InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse],
        ) -> InvokeHandler[ConfigSubmitInvokeActivity, ConfigInvokeResponse]:
            validate_handler_type(
                self.logger, func, ConfigSubmitInvokeActivity, "on_config_submit", "ConfigSubmitInvokeActivity"
            )
            config = ACTIVITY_ROUTES["config.submit"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_file_consent(
        self, handler: VoidInvokeHandler[FileConsentInvokeActivity]
    ) -> VoidInvokeHandler[FileConsentInvokeActivity]: ...

    @overload
    def on_file_consent(
        self,
    ) -> Callable[[VoidInvokeHandler[FileConsentInvokeActivity]], VoidInvokeHandler[FileConsentInvokeActivity]]: ...

    def on_file_consent(
        self, handler: Optional[VoidInvokeHandler[FileConsentInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[FileConsentInvokeActivity]:
        """Register a file.consent activity handler."""

        def decorator(
            func: VoidInvokeHandler[FileConsentInvokeActivity],
        ) -> VoidInvokeHandler[FileConsentInvokeActivity]:
            validate_handler_type(
                self.logger, func, FileConsentInvokeActivity, "on_file_consent", "FileConsentInvokeActivity"
            )
            config = ACTIVITY_ROUTES["file.consent"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_execute(
        self, handler: VoidInvokeHandler[ExecuteActionInvokeActivity]
    ) -> VoidInvokeHandler[ExecuteActionInvokeActivity]: ...

    @overload
    def on_message_execute(
        self,
    ) -> Callable[[VoidInvokeHandler[ExecuteActionInvokeActivity]], VoidInvokeHandler[ExecuteActionInvokeActivity]]: ...

    def on_message_execute(
        self, handler: Optional[VoidInvokeHandler[ExecuteActionInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[ExecuteActionInvokeActivity]:
        """Register a message.execute activity handler."""

        def decorator(
            func: VoidInvokeHandler[ExecuteActionInvokeActivity],
        ) -> VoidInvokeHandler[ExecuteActionInvokeActivity]:
            validate_handler_type(
                self.logger, func, ExecuteActionInvokeActivity, "on_message_execute", "ExecuteActionInvokeActivity"
            )
            config = ACTIVITY_ROUTES["message.execute"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_query_link(
        self, handler: InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]: ...

    @overload
    def on_message_ext_query_link(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]],
        InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse],
    ]: ...

    def on_message_ext_query_link(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]:
        """Register a message.ext.query-link activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionQueryLinkInvokeActivity,
                "on_message_ext_query_link",
                "MessageExtensionQueryLinkInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.query-link"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_anon_query_link(
        self, handler: InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]: ...

    @overload
    def on_message_ext_anon_query_link(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]],
        InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse],
    ]: ...

    def on_message_ext_anon_query_link(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]:
        """Register a message.ext.anon-query-link activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionAnonQueryLinkInvokeActivity, MessagingExtensionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionAnonQueryLinkInvokeActivity,
                "on_message_ext_anon_query_link",
                "MessageExtensionAnonQueryLinkInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.anon-query-link"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_query(
        self, handler: InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse]: ...

    @overload
    def on_message_ext_query(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse]],
        InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse],
    ]: ...

    def on_message_ext_query(
        self,
        handler: Optional[InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse]] = None,
    ) -> InvokeHandlerUnion[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse]:
        """Register a message.ext.query activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionQueryInvokeActivity, MessagingExtensionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionQueryInvokeActivity,
                "on_message_ext_query",
                "MessageExtensionQueryInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.query"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_select_item(
        self, handler: InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse]: ...

    @overload
    def on_message_ext_select_item(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse]],
        InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse],
    ]: ...

    def on_message_ext_select_item(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse]:
        """Register a message.ext.select-item activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionSelectItemInvokeActivity, MessagingExtensionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionSelectItemInvokeActivity,
                "on_message_ext_select_item",
                "MessageExtensionSelectItemInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.select-item"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_submit(
        self, handler: InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse]: ...

    @overload
    def on_message_ext_submit(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse]],
        InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse],
    ]: ...

    def on_message_ext_submit(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse]:
        """Register a message.ext.submit activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionSubmitActionInvokeActivity, MessagingExtensionActionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionSubmitActionInvokeActivity,
                "on_message_ext_submit",
                "MessageExtensionSubmitActionInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.submit"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_open(
        self, handler: InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse]: ...

    @overload
    def on_message_ext_open(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse]],
        InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse],
    ]: ...

    def on_message_ext_open(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse]:
        """Register a message.ext.open activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionFetchTaskInvokeActivity, MessagingExtensionActionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionFetchTaskInvokeActivity,
                "on_message_ext_open",
                "MessageExtensionFetchTaskInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.open"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_query_settings_url(
        self, handler: InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse]: ...

    @overload
    def on_message_ext_query_settings_url(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse]],
        InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse],
    ]: ...

    def on_message_ext_query_settings_url(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse]:
        """Register a message.ext.query-settings-url activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionQuerySettingUrlInvokeActivity, MessagingExtensionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionQuerySettingUrlInvokeActivity,
                "on_message_ext_query_settings_url",
                "MessageExtensionQuerySettingUrlInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.query-settings-url"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_setting(
        self, handler: InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse]
    ) -> InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse]: ...

    @overload
    def on_message_ext_setting(
        self,
    ) -> Callable[
        [InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse]],
        InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse],
    ]: ...

    def on_message_ext_setting(
        self,
        handler: Optional[
            InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse]
        ] = None,
    ) -> InvokeHandlerUnion[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse]:
        """Register a message.ext.setting activity handler."""

        def decorator(
            func: InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse],
        ) -> InvokeHandler[MessageExtensionSettingInvokeActivity, MessagingExtensionInvokeResponse]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionSettingInvokeActivity,
                "on_message_ext_setting",
                "MessageExtensionSettingInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.setting"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_ext_card_button_clicked(
        self, handler: VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity]
    ) -> VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity]: ...

    @overload
    def on_message_ext_card_button_clicked(
        self,
    ) -> Callable[
        [VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity]],
        VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity],
    ]: ...

    def on_message_ext_card_button_clicked(
        self, handler: Optional[VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[MessageExtensionCardButtonClickedInvokeActivity]:
        """Register a message.ext.card-button-clicked activity handler."""

        def decorator(
            func: VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity],
        ) -> VoidInvokeHandler[MessageExtensionCardButtonClickedInvokeActivity]:
            validate_handler_type(
                self.logger,
                func,
                MessageExtensionCardButtonClickedInvokeActivity,
                "on_message_ext_card_button_clicked",
                "MessageExtensionCardButtonClickedInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.ext.card-button-clicked"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_dialog_open(
        self, handler: InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse]
    ) -> InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse]: ...

    @overload
    def on_dialog_open(
        self,
    ) -> Callable[
        [InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse]],
        InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse],
    ]: ...

    def on_dialog_open(
        self, handler: Optional[InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse]] = None
    ) -> InvokeHandlerUnion[TaskFetchInvokeActivity, TaskModuleInvokeResponse]:
        """Register a dialog.open activity handler."""

        def decorator(
            func: InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse],
        ) -> InvokeHandler[TaskFetchInvokeActivity, TaskModuleInvokeResponse]:
            validate_handler_type(
                self.logger, func, TaskFetchInvokeActivity, "on_dialog_open", "TaskFetchInvokeActivity"
            )
            config = ACTIVITY_ROUTES["dialog.open"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_dialog_submit(
        self, handler: InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse]
    ) -> InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse]: ...

    @overload
    def on_dialog_submit(
        self,
    ) -> Callable[
        [InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse]],
        InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse],
    ]: ...

    def on_dialog_submit(
        self, handler: Optional[InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse]] = None
    ) -> InvokeHandlerUnion[TaskSubmitInvokeActivity, TaskModuleInvokeResponse]:
        """Register a dialog.submit activity handler."""

        def decorator(
            func: InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse],
        ) -> InvokeHandler[TaskSubmitInvokeActivity, TaskModuleInvokeResponse]:
            validate_handler_type(
                self.logger, func, TaskSubmitInvokeActivity, "on_dialog_submit", "TaskSubmitInvokeActivity"
            )
            config = ACTIVITY_ROUTES["dialog.submit"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_tab_open(
        self, handler: InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse]
    ) -> InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse]: ...

    @overload
    def on_tab_open(
        self,
    ) -> Callable[
        [InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse]],
        InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse],
    ]: ...

    def on_tab_open(
        self, handler: Optional[InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse]] = None
    ) -> InvokeHandlerUnion[TabFetchInvokeActivity, TabInvokeResponse]:
        """Register a tab.open activity handler."""

        def decorator(
            func: InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse],
        ) -> InvokeHandler[TabFetchInvokeActivity, TabInvokeResponse]:
            validate_handler_type(self.logger, func, TabFetchInvokeActivity, "on_tab_open", "TabFetchInvokeActivity")
            config = ACTIVITY_ROUTES["tab.open"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_tab_submit(
        self, handler: InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse]
    ) -> InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse]: ...

    @overload
    def on_tab_submit(
        self,
    ) -> Callable[
        [InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse]],
        InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse],
    ]: ...

    def on_tab_submit(
        self, handler: Optional[InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse]] = None
    ) -> InvokeHandlerUnion[TabSubmitInvokeActivity, TabInvokeResponse]:
        """Register a tab.submit activity handler."""

        def decorator(
            func: InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse],
        ) -> InvokeHandler[TabSubmitInvokeActivity, TabInvokeResponse]:
            validate_handler_type(
                self.logger, func, TabSubmitInvokeActivity, "on_tab_submit", "TabSubmitInvokeActivity"
            )
            config = ACTIVITY_ROUTES["tab.submit"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_submit(
        self, handler: VoidInvokeHandler[MessageSubmitActionInvokeActivity]
    ) -> VoidInvokeHandler[MessageSubmitActionInvokeActivity]: ...

    @overload
    def on_message_submit(
        self,
    ) -> Callable[
        [VoidInvokeHandler[MessageSubmitActionInvokeActivity]], VoidInvokeHandler[MessageSubmitActionInvokeActivity]
    ]: ...

    def on_message_submit(
        self, handler: Optional[VoidInvokeHandler[MessageSubmitActionInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[MessageSubmitActionInvokeActivity]:
        """Register a message.submit activity handler."""

        def decorator(
            func: VoidInvokeHandler[MessageSubmitActionInvokeActivity],
        ) -> VoidInvokeHandler[MessageSubmitActionInvokeActivity]:
            validate_handler_type(
                self.logger,
                func,
                MessageSubmitActionInvokeActivity,
                "on_message_submit",
                "MessageSubmitActionInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.submit"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_message_submit_feedback(
        self, handler: VoidInvokeHandler[MessageSubmitActionInvokeActivity]
    ) -> VoidInvokeHandler[MessageSubmitActionInvokeActivity]: ...

    @overload
    def on_message_submit_feedback(
        self,
    ) -> Callable[
        [VoidInvokeHandler[MessageSubmitActionInvokeActivity]], VoidInvokeHandler[MessageSubmitActionInvokeActivity]
    ]: ...

    def on_message_submit_feedback(
        self, handler: Optional[VoidInvokeHandler[MessageSubmitActionInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[MessageSubmitActionInvokeActivity]:
        """Register a message.submit.feedback activity handler."""

        def decorator(
            func: VoidInvokeHandler[MessageSubmitActionInvokeActivity],
        ) -> VoidInvokeHandler[MessageSubmitActionInvokeActivity]:
            validate_handler_type(
                self.logger,
                func,
                MessageSubmitActionInvokeActivity,
                "on_message_submit_feedback",
                "MessageSubmitActionInvokeActivity",
            )
            config = ACTIVITY_ROUTES["message.submit.feedback"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_handoff_action(
        self, handler: VoidInvokeHandler[HandoffActionInvokeActivity]
    ) -> VoidInvokeHandler[HandoffActionInvokeActivity]: ...

    @overload
    def on_handoff_action(
        self,
    ) -> Callable[[VoidInvokeHandler[HandoffActionInvokeActivity]], VoidInvokeHandler[HandoffActionInvokeActivity]]: ...

    def on_handoff_action(
        self, handler: Optional[VoidInvokeHandler[HandoffActionInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[HandoffActionInvokeActivity]:
        """Register a handoff.action activity handler."""

        def decorator(
            func: VoidInvokeHandler[HandoffActionInvokeActivity],
        ) -> VoidInvokeHandler[HandoffActionInvokeActivity]:
            validate_handler_type(
                self.logger, func, HandoffActionInvokeActivity, "on_handoff_action", "HandoffActionInvokeActivity"
            )
            config = ACTIVITY_ROUTES["handoff.action"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_signin_token_exchange(
        self, handler: InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType]
    ) -> InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType]: ...

    @overload
    def on_signin_token_exchange(
        self,
    ) -> Callable[
        [InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType]],
        InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType],
    ]: ...

    def on_signin_token_exchange(
        self,
        handler: Optional[InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType]] = None,
    ) -> InvokeHandlerUnion[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType]:
        """Register a signin.token-exchange activity handler."""

        def decorator(
            func: InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType],
        ) -> InvokeHandler[SignInTokenExchangeInvokeActivity, TokenExchangeInvokeResponseType]:
            validate_handler_type(
                self.logger,
                func,
                SignInTokenExchangeInvokeActivity,
                "on_signin_token_exchange",
                "SignInTokenExchangeInvokeActivity",
            )
            config = ACTIVITY_ROUTES["signin.token-exchange"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_signin_verify_state(
        self, handler: VoidInvokeHandler[SignInVerifyStateInvokeActivity]
    ) -> VoidInvokeHandler[SignInVerifyStateInvokeActivity]: ...

    @overload
    def on_signin_verify_state(
        self,
    ) -> Callable[
        [VoidInvokeHandler[SignInVerifyStateInvokeActivity]], VoidInvokeHandler[SignInVerifyStateInvokeActivity]
    ]: ...

    def on_signin_verify_state(
        self, handler: Optional[VoidInvokeHandler[SignInVerifyStateInvokeActivity]] = None
    ) -> VoidInvokeHandlerUnion[SignInVerifyStateInvokeActivity]:
        """Register a signin.verify-state activity handler."""

        def decorator(
            func: VoidInvokeHandler[SignInVerifyStateInvokeActivity],
        ) -> VoidInvokeHandler[SignInVerifyStateInvokeActivity]:
            validate_handler_type(
                self.logger,
                func,
                SignInVerifyStateInvokeActivity,
                "on_signin_verify_state",
                "SignInVerifyStateInvokeActivity",
            )
            config = ACTIVITY_ROUTES["signin.verify-state"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_card_action(
        self, handler: InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse]
    ) -> InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse]: ...

    @overload
    def on_card_action(
        self,
    ) -> Callable[
        [InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse]],
        InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse],
    ]: ...

    def on_card_action(
        self, handler: Optional[InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse]] = None
    ) -> InvokeHandlerUnion[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse]:
        """Register a card.action activity handler."""

        def decorator(
            func: InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse],
        ) -> InvokeHandler[AdaptiveCardInvokeActivity, AdaptiveCardInvokeResponse]:
            validate_handler_type(
                self.logger, func, AdaptiveCardInvokeActivity, "on_card_action", "AdaptiveCardInvokeActivity"
            )
            config = ACTIVITY_ROUTES["card.action"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_invoke(self, handler: BasicHandler[InvokeActivity]) -> BasicHandler[InvokeActivity]: ...

    @overload
    def on_invoke(self) -> Callable[[BasicHandler[InvokeActivity]], BasicHandler[InvokeActivity]]: ...

    def on_invoke(self, handler: Optional[BasicHandler[InvokeActivity]] = None) -> BasicHandlerUnion[InvokeActivity]:
        """Register a invoke activity handler."""

        def decorator(func: BasicHandler[InvokeActivity]) -> BasicHandler[InvokeActivity]:
            validate_handler_type(self.logger, func, InvokeActivity, "on_invoke", "InvokeActivity")
            config = ACTIVITY_ROUTES["invoke"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_installation_update(
        self, handler: BasicHandler[InstallUpdateActivity]
    ) -> BasicHandler[InstallUpdateActivity]: ...

    @overload
    def on_installation_update(
        self,
    ) -> Callable[[BasicHandler[InstallUpdateActivity]], BasicHandler[InstallUpdateActivity]]: ...

    def on_installation_update(
        self, handler: Optional[BasicHandler[InstallUpdateActivity]] = None
    ) -> BasicHandlerUnion[InstallUpdateActivity]:
        """Register a installation_update activity handler."""

        def decorator(func: BasicHandler[InstallUpdateActivity]) -> BasicHandler[InstallUpdateActivity]:
            validate_handler_type(
                self.logger, func, InstallUpdateActivity, "on_installation_update", "InstallUpdateActivity"
            )
            config = ACTIVITY_ROUTES["installation_update"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_install_add(self, handler: BasicHandler[InstalledActivity]) -> BasicHandler[InstalledActivity]: ...

    @overload
    def on_install_add(self) -> Callable[[BasicHandler[InstalledActivity]], BasicHandler[InstalledActivity]]: ...

    def on_install_add(
        self, handler: Optional[BasicHandler[InstalledActivity]] = None
    ) -> BasicHandlerUnion[InstalledActivity]:
        """Register a install.add activity handler."""

        def decorator(func: BasicHandler[InstalledActivity]) -> BasicHandler[InstalledActivity]:
            validate_handler_type(self.logger, func, InstalledActivity, "on_install_add", "InstalledActivity")
            config = ACTIVITY_ROUTES["install.add"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_install_remove(self, handler: BasicHandler[UninstalledActivity]) -> BasicHandler[UninstalledActivity]: ...

    @overload
    def on_install_remove(self) -> Callable[[BasicHandler[UninstalledActivity]], BasicHandler[UninstalledActivity]]: ...

    def on_install_remove(
        self, handler: Optional[BasicHandler[UninstalledActivity]] = None
    ) -> BasicHandlerUnion[UninstalledActivity]:
        """Register a install.remove activity handler."""

        def decorator(func: BasicHandler[UninstalledActivity]) -> BasicHandler[UninstalledActivity]:
            validate_handler_type(self.logger, func, UninstalledActivity, "on_install_remove", "UninstalledActivity")
            config = ACTIVITY_ROUTES["install.remove"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_typing(self, handler: BasicHandler[TypingActivity]) -> BasicHandler[TypingActivity]: ...

    @overload
    def on_typing(self) -> Callable[[BasicHandler[TypingActivity]], BasicHandler[TypingActivity]]: ...

    def on_typing(self, handler: Optional[BasicHandler[TypingActivity]] = None) -> BasicHandlerUnion[TypingActivity]:
        """Register a typing activity handler."""

        def decorator(func: BasicHandler[TypingActivity]) -> BasicHandler[TypingActivity]:
            validate_handler_type(self.logger, func, TypingActivity, "on_typing", "TypingActivity")
            config = ACTIVITY_ROUTES["typing"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_trace(self, handler: BasicHandler[TraceActivity]) -> BasicHandler[TraceActivity]: ...

    @overload
    def on_trace(self) -> Callable[[BasicHandler[TraceActivity]], BasicHandler[TraceActivity]]: ...

    def on_trace(self, handler: Optional[BasicHandler[TraceActivity]] = None) -> BasicHandlerUnion[TraceActivity]:
        """Register a trace activity handler."""

        def decorator(func: BasicHandler[TraceActivity]) -> BasicHandler[TraceActivity]:
            validate_handler_type(self.logger, func, TraceActivity, "on_trace", "TraceActivity")
            config = ACTIVITY_ROUTES["trace"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_handoff(self, handler: BasicHandler[HandoffActivity]) -> BasicHandler[HandoffActivity]: ...

    @overload
    def on_handoff(self) -> Callable[[BasicHandler[HandoffActivity]], BasicHandler[HandoffActivity]]: ...

    def on_handoff(self, handler: Optional[BasicHandler[HandoffActivity]] = None) -> BasicHandlerUnion[HandoffActivity]:
        """Register a handoff activity handler."""

        def decorator(func: BasicHandler[HandoffActivity]) -> BasicHandler[HandoffActivity]:
            validate_handler_type(self.logger, func, HandoffActivity, "on_handoff", "HandoffActivity")
            config = ACTIVITY_ROUTES["handoff"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator

    @overload
    def on_activity(self, handler: BasicHandler[Activity]) -> BasicHandler[Activity]: ...

    @overload
    def on_activity(self) -> Callable[[BasicHandler[Activity]], BasicHandler[Activity]]: ...

    def on_activity(self, handler: Optional[BasicHandler[Activity]] = None) -> BasicHandlerUnion[Activity]:
        """Register a activity activity handler."""

        def decorator(func: BasicHandler[Activity]) -> BasicHandler[Activity]:
            validate_handler_type(self.logger, func, Activity, "on_activity", "Activity")
            config = ACTIVITY_ROUTES["activity"]
            self.router.add_handler(config.selector, func)
            return func

        if handler is not None:
            return decorator(handler)
        return decorator



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from .jwt_middleware import create_jwt_validation_middleware
from .remote_function_jwt_middleware import remote_function_jwt_validation
from .token_validator import TokenValidator

__all__ = ["TokenValidator", "create_jwt_validation_middleware", "remote_function_jwt_validation"]



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""
# pyright: basic

import asyncio
from datetime import datetime, timedelta
from unittest.mock import MagicMock

import pytest
from microsoft.teams.api import (
    Account,
    ApiClient,
    ConversationAccount,
    ConversationReference,
    SentActivity,
    TypingActivityInput,
)
from microsoft.teams.apps import HttpStream


@pytest.mark.skip(reason="introduces delays in CI pipeline")
class TestHttpStream:
    @pytest.fixture
    def mock_logger(self):
        return MagicMock()

    @pytest.fixture
    def mock_api_client(self):
        client = MagicMock(spec=ApiClient)

        mock_activities = MagicMock()
        mock_conversations = MagicMock()
        mock_conversations.activities.return_value = mock_activities
        client.conversations = mock_conversations

        client.send_call_count = 0
        client.send_times = []
        client.sent_activities = []

        async def mock_send(activity):
            client.send_call_count += 1
            client.send_times.append(datetime.now())
            client.sent_activities.append(activity)
            return SentActivity(id=f"test-id-{client.send_call_count}", activity_params=activity)

        client.conversations.activities().create = mock_send

        return client

    @pytest.fixture
    def conversation_reference(self):
        return ConversationReference(
            service_url="https://smba.trafficmanager.net/teams/",
            bot=Account(id="test-bot", name="Test Bot"),
            conversation=ConversationAccount(id="test-conversation", conversation_type="personal"),
            activity_id="test-activity",
            channel_id="msteams",
        )

    @pytest.fixture
    def http_stream(self, mock_api_client, conversation_reference, mock_logger):
        return HttpStream(mock_api_client, conversation_reference, mock_logger)

    @pytest.mark.asyncio
    async def test_stream_emit_message_flushes_after_500ms(self, mock_api_client, conversation_reference, mock_logger):
        """Test that messages are flushed after 500ms timeout."""

        stream = HttpStream(mock_api_client, conversation_reference, mock_logger)
        start_time = datetime.now()

        stream.emit("Test message")
        await asyncio.sleep(0.6)  # Wait longer than 500ms timeout

        assert mock_api_client.send_call_count > 0, "Should have sent at least one message"
        assert any(t >= start_time + timedelta(milliseconds=450) for t in mock_api_client.send_times), (
            "Should have waited approximately 500ms before sending"
        )

    @pytest.mark.asyncio
    async def test_stream_multiple_emits_restarts_timer(self, mock_api_client, conversation_reference, mock_logger):
        """Test that multiple emits reset the timer."""

        stream = HttpStream(mock_api_client, conversation_reference, mock_logger)

        stream.emit("First message")
        await asyncio.sleep(0.3)  # Wait less than 500ms

        stream.emit("Second message")  # This should reset the timer
        await asyncio.sleep(0.3)  # Still less than 500ms from second emit
        assert mock_api_client.send_call_count == 0, "Should not have sent yet"
        await asyncio.sleep(0.3)  # Now over 500ms from second emit
        assert mock_api_client.send_call_count > 0, "Should have sent messages after timer expired"

    @pytest.mark.asyncio
    async def test_stream_send_timeout_handled_gracefully(self, mock_api_client, conversation_reference, mock_logger):
        """Test that send timeouts are handled gracefully with retries."""
        call_count = 0

        async def mock_send_with_timeout(activity):
            nonlocal call_count
            call_count += 1
            if call_count == 1:  # Fail first attempt
                raise TimeoutError("Operation timed out")

            # Succeed on second attempt
            return SentActivity(id=f"success-after-timeout-{call_count}", activity_params=activity)

        mock_api_client.conversations.activities().create = mock_send_with_timeout

        stream = HttpStream(mock_api_client, conversation_reference, mock_logger)

        stream.emit("Test message with timeout")
        await asyncio.sleep(0.8)  # Wait for flush and retries

        result = await stream.close()

        assert call_count > 1, "Should have retried after timeout"
        assert result is not None

    @pytest.mark.asyncio
    async def test_stream_all_timeouts_fail_handled_gracefully(
        self, mock_api_client, conversation_reference, mock_logger
    ):
        """Test that when all timeouts fail, it's handled gracefully."""
        call_count = 0

        async def mock_send_all_timeout(activity):
            nonlocal call_count
            call_count += 1
            raise TimeoutError("All operations timed out")

        mock_api_client.conversations.activities().create = mock_send_all_timeout

        stream = HttpStream(mock_api_client, conversation_reference, mock_logger)

        stream.emit("Test message with all timeouts")
        await asyncio.sleep(5.0)  # Wait for flush and all retries to complete

        await stream.close()
        assert call_count > 1, "Should have retried after timeout"

    @pytest.mark.asyncio
    async def test_stream_update_status_sends_typing_activity(
        self, mock_api_client, conversation_reference, mock_logger
    ):
        """Test that update sends typing activities."""
        stream = HttpStream(mock_api_client, conversation_reference, mock_logger)

        stream.update("Thinking...")
        await asyncio.sleep(0.6)  # Wait for the flush task to complete

        assert stream.count > 0 or len(mock_api_client.sent_activities) > 0, "Should have processed the update"
        assert stream.sequence >= 2, "Should increment sequence after sending"

        assert len(mock_api_client.sent_activities) > 0, "Should have sent at least one activity"
        sent_activity = mock_api_client.sent_activities[0]
        assert isinstance(sent_activity, TypingActivityInput)
        assert sent_activity.text == "Thinking..."
        assert sent_activity.channel_data is not None
        assert sent_activity.channel_data.stream_type == "informative"

    @pytest.mark.asyncio
    async def test_stream_sequence_of_update_and_emit(self, mock_api_client, conversation_reference, mock_logger):
        """Test a sequence of update() followed by emit(), ensuring correct ordering and flush behavior."""

        stream = HttpStream(mock_api_client, conversation_reference, mock_logger)

        stream.update("Preparing response...")
        await asyncio.sleep(0.6)

        stream.emit("Final response message")
        await asyncio.sleep(0.6)

        assert len(mock_api_client.sent_activities) >= 2, "Should have sent typing activity and message"

        typing_activity = mock_api_client.sent_activities[0]
        message_activity = mock_api_client.sent_activities[1]

        # First should be typing activity from update()
        assert isinstance(typing_activity, TypingActivityInput)
        assert typing_activity.text == "Preparing response..."

        # Second should be a normal message from emit()
        assert message_activity.text == "Final response message"

        # Sequence numbers should have increased
        assert stream.sequence >= 3, "Sequence should increment for both update and emit"



"""
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the MIT License.
"""

from logging import Logger
from typing import Any, Awaitable, Callable, Dict, List, Optional

from fastapi import HTTPException, Request, Response

from ..contexts import ClientContext
from .token_validator import TokenValidator


def require_fields(fields: Dict[str, Optional[Any]], context: str, logger: Logger) -> None:
    missing: List[str] = [name for name, value in fields.items() if not value]
    if missing:
        message = f"Missing or invalid fields in {context}: {', '.join(missing)}"
        logger.warning(message)
        raise HTTPException(status_code=401, detail=message)


def remote_function_jwt_validation(logger: Logger, entra_token_validator: Optional[TokenValidator]):
    """
    Middleware to validate JWT for remote function calls.
    Args:
        entra_token_validator: TokenValidator instance for Entra ID tokens
        logger: Logger instance

    Returns:
        Middleware function that can be added to FastAPI app
    """

    async def middleware(request: Request, call_next: Callable[[Request], Awaitable[Response]]) -> Response:
        # Extract auth token
        authorization = request.headers.get("Authorization", "")
        parts = authorization.split(" ")
        auth_token = parts[1] if len(parts) == 2 and parts[0].lower() == "bearer" else ""

        # Validate headers
        require_fields(
            {
                "X-Teams-App-Session-Id": request.headers.get("X-Teams-App-Session-Id"),
                "X-Teams-Page-Id": request.headers.get("X-Teams-Page-Id"),
                "Authorization (Bearer token)": auth_token,
            },
            "header",
            logger,
        )

        if not entra_token_validator:
            raise HTTPException(status_code=500, detail="Token validator not configured")

        # Validate token
        token_payload = await entra_token_validator.validate_token(auth_token)

        # Validate required fields in token
        require_fields(
            {"oid": token_payload.get("oid"), "tid": token_payload.get("tid"), "name": token_payload.get("name")},
            "token payload",
            logger,
        )

        # Build context
        request.state.context = ClientContext(
            app_session_id=request.headers.get("X-Teams-App-Session-Id"),  # type: ignore
            tenant_id=token_payload["tid"],
            user_id=token_payload["oid"],
            user_name=token_payload["name"],
            page_id=request.headers.get("X-Teams-Page-Id"),  # type: ignore
            auth_token=auth_token,  # type: ignore
            app_id=token_payload.get("appId"),
            channel_id=request.headers.get("X-Teams-Channel-Id"),
            chat_id=request.headers.get("X-Teams-Chat-Id"),
            meeting_id=request.headers.get("X-Teams-Meeting-Id"),
            message_id=request.headers.get("X-Teams-Message-Id"),
            sub_page_id=request.headers.get("X-Teams-Sub-Page-Id"),
            team_id=request.headers.get("X-Teams-Team-Id"),
        )
        return await call_next(request)

    return middleware
